#include "addr.h"
#include "linux_consts.h"
#include <xenctrl.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "htable.h"

#include <xen/sys/privcmd.h>

#ifndef MAP_ALL

#define L1_PAGETABLE_SHIFT      12
#define L2_PAGETABLE_SHIFT      21
#define L3_PAGETABLE_SHIFT      30
#define L4_PAGETABLE_SHIFT      39

#define L1_PAGETABLE_ENTRIES    512
#define L2_PAGETABLE_ENTRIES    512
#define L3_PAGETABLE_ENTRIES    512
#define L4_PAGETABLE_ENTRIES    512


#define l1_table_offset(_a) \
  (((_a) >> L1_PAGETABLE_SHIFT) & (L1_PAGETABLE_ENTRIES - 1))
#define l2_table_offset(_a) \
  (((_a) >> L2_PAGETABLE_SHIFT) & (L2_PAGETABLE_ENTRIES - 1))
#define l3_table_offset(_a) \
   (((_a) >> L3_PAGETABLE_SHIFT) & (L3_PAGETABLE_ENTRIES - 1))
#define l4_table_offset(_a) \
    (((_a) >> L4_PAGETABLE_SHIFT) & (L4_PAGETABLE_ENTRIES - 1))

#define _PAGE_PRESENT  0x001UL

//#define PAGE_SHIFT L1_PAGETABLE_SHIFT

//#define to_ma(cpu, maddr) ({ (maddr); })

static inline uint64_t to_ma(int cpu, uint64_t maddr)
{
  return maddr;
}

static struct htable *ht = NULL;

static unsigned long
my_xc_translate_foreign_address(XCH xc_handle, uint32_t dom, int cpu, void* guest_va)
{
  uint32_t current_domid = dom;
  
  unsigned long l4e, l3e, l2e, l1e, l3p, l2p, l1p, p, va = (unsigned long)guest_va;
  uint64_t *l4, *l3, *l2, *l1;

  vcpu_guest_context_any_t ctxt;
  int n;

  unsigned long mfn;

  if (ht == NULL)
      ht = htable_new();

  mfn = htable_get(ht, va & ~(PAGE_SIZE - 1));
  if (mfn != 0)
      return mfn;

  if ((n = xc_vcpu_getcontext(xc_handle, dom, cpu, &ctxt)) < 0) {
    perror("xc_vcpu_getcontext");
    return n;
  }

  assert((ctxt.x64.ctrlreg[4] & 0x20) != 0); /* must not be legacy ia32 mode */

  l4 = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, xen_cr3_to_pfn(ctxt.x64.ctrlreg[3]));
  
  if ( l4 == NULL )
    return -1;

  l4e = l4[l4_table_offset(va)];
  munmap(l4, PAGE_SIZE);
  if ( !(l4e & _PAGE_PRESENT) )
    return -1;
  l3p = to_ma(cpu, l4e);
  l3 = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, l3p >> PAGE_SHIFT);
  if ( l3 == NULL )
    return -1;
  
  l3e = l3[l3_table_offset(va)];
  munmap(l3, PAGE_SIZE);
  if ( !(l3e & _PAGE_PRESENT) )
    return -1;
  l2p = to_ma(cpu, l3e);
  l2 = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, l2p >> PAGE_SHIFT);
  if ( l2 == NULL )
    return -1;
  
  l2e = l2[l2_table_offset(va)];
  munmap(l2, PAGE_SIZE);
  if ( !(l2e & _PAGE_PRESENT) )
    return -1;
  l1p = to_ma(cpu, l2e);
  if (l2e & 0x80) { /* 2M pages */
    p = to_ma(cpu, l1p + (l1_table_offset(va) << PAGE_SHIFT));
  } else { /* 4K pages */
    l1 = xc_map_foreign_range(xc_handle, current_domid, PAGE_SIZE, PROT_READ, l1p >> PAGE_SHIFT);
    if ( l1 == NULL )
      return -1;
    
    l1e = l1[l1_table_offset(va)];
    munmap(l1, PAGE_SIZE);
    if ( !(l1e & _PAGE_PRESENT) )
      return -1;
    p = to_ma(cpu, l1e);
  }

  mfn =  p >> PAGE_SHIFT;

  htable_put(ht, va & ~(PAGE_SIZE - 1), mfn);

  return mfn;
}

inline void*
addr_trans(XCH xc_handle, int dom, int cpu, uint32_t flag, unsigned long addr,
           unsigned int size)
{
    unsigned long mfn, va, ga;
    xen_pfn_t *arr;
    int n;

    arr = (xen_pfn_t *)malloc(sizeof(xen_pfn_t) * (size / PAGE_SIZE + 1));

    ga = addr & ~(PAGE_SIZE - 1);
    n = 0;

    do {
        mfn = my_xc_translate_foreign_address(xc_handle, dom, cpu, (void*)ga);
        if (mfn == -1) {
            printf("addr_trans: translation error (%lx)\n", ga);
            free(arr);
            return NULL;
        }

        arr[n] = mfn;

        ga += PAGE_SIZE;
        n++;
    } while (ga < addr + size);

    va = (unsigned long)xc_map_foreign_batch(xc_handle, dom, flag, arr, n);
    if (va == 0) {
        printf("addr_trans: map error (%lx)\n", addr);
        return NULL;
    }

    free(arr);

    return (void*)((unsigned long)va +
                   ((unsigned long)addr & (PAGE_SIZE - 1)));
}

void unmap_pages(void *addr, unsigned int size)
{
    unsigned long ga, start;
    unsigned int real_size;

    ga = (unsigned long)addr;
    start = ga & ~(PAGE_SIZE - 1);
    real_size = (size + (ga - start) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1);

    munmap((void *)start, real_size);
}

#else /* MAP_ALL */

static int get_mfn_list(XCH xc_handle, unsigned long max_extents,
                        xen_pfn_t *extent_start)
{
#ifdef XEN41
    return xc_machphys_mfn_list(xc_handle, max_extents, extent_start);
#else
    struct xen_machphys_mfn_list xmml = {
        .max_extents = max_extents,
    };
    
    set_xen_guest_handle(xmml.extent_start, extent_start);

    if (xc_memory_op(xc_handle, XENMEM_machphys_mfn_list, &xmml) ||
        (xmml.nr_extents != max_extents)) {
        fprintf(stderr, "xc_get_m2p_mfns\n");
        return -1;
    }

    return 0;
#endif
}

#define ROUNDUP(_x,_w) (((unsigned long)(_x)+(1UL<<(_w))-1) & ~((1UL<<(_w))-1))

#define L2_PAGETABLE_SHIFT_PAE        21
#define M2P_SHIFT       L2_PAGETABLE_SHIFT_PAE
#define M2P_CHUNK_SIZE  (1 << M2P_SHIFT)
#define M2P_SIZE(_m)    ROUNDUP(((_m) * sizeof(xen_pfn_t)), M2P_SHIFT)
#define M2P_CHUNKS(_m)  (M2P_SIZE((_m)) >> M2P_SHIFT)

void *xc_map_foreign_ranges(XCH xc_handle, uint32_t dom,
                            size_t size, int prot, size_t chunksize,
                            privcmd_mmap_entry_t entries[], int nentries);

static xen_pfn_t *m2p = NULL;

static xen_pfn_t *get_m2p_table(XCH xc_handle)
{
    privcmd_mmap_entry_t *entries;
    unsigned long m2p_chunks, m2p_size;
    xen_pfn_t *extent_start;
    unsigned long max_mfn;
    int i;

#ifdef XEN41
    max_mfn = xc_maximum_ram_page(xc_handle);
#else
    max_mfn = xc_memory_op(xc_handle, XENMEM_maximum_ram_page, NULL);
#endif

    m2p_size = M2P_SIZE(max_mfn);
    m2p_chunks = M2P_CHUNKS(max_mfn);

    extent_start = calloc(m2p_chunks, sizeof(xen_pfn_t));
    if (!extent_start) {
        fprintf(stderr, "failed to allocate space for m2p mfns\n");
        goto err0;
    }

    if (get_mfn_list(xc_handle, m2p_chunks, extent_start)) {
        fprintf(stderr, "get_m2p_mfns\n");
        goto err1;
    }

    entries = calloc(m2p_chunks, sizeof(privcmd_mmap_entry_t));
    if (entries == NULL) {
        fprintf(stderr, "failed to allocate space for mmap entries\n");
        goto err1;
    }

    for (i = 0; i < m2p_chunks; i++)
        entries[i].mfn = extent_start[i];

    m2p = xc_map_foreign_ranges(xc_handle, DOMID_XEN,
                                m2p_size, PROT_READ, M2P_CHUNK_SIZE,
                                entries, m2p_chunks);
    if (m2p == NULL) {
        fprintf(stderr, "xc_mmap_foreign_ranges failed\n");
        goto err2;
    }

err2:
    free(entries);
err1:
    free(extent_start);

err0:
    return m2p;
}

static int get_maximum_gpfn(XCH xc_handle, int dom)
{
#ifdef XEN41
    return xc_domain_maximum_gpfn(xc_handle, dom);
#else
    return xc_memory_op(xc_handle, XENMEM_maximum_gpfn, &dom);
#endif
}

static void *mem_base = NULL;
static unsigned long cr3;
static int hvm;

void addr_init(XCH xc_handle, int dom, int cpu)
{
    unsigned long nr_mfns, p2m_size;
    xen_pfn_t *p2m;
    int *err;
    xen_pfn_t *mfn_list;
    unsigned long mfn, pfn;
    vcpu_guest_context_any_t ctxt;
    xc_dominfo_t info;
    int i;

    if (xc_domain_getinfo(xc_handle, dom, 1, &info) < 0) {
        perror("xc_domain_getinfo");
        return;
    }

    hvm = info.hvm;

    if (xc_vcpu_getcontext(xc_handle, dom, cpu, &ctxt) < 0) {
        perror("xc_vcpu_getcontext");
        return;
    }

    cr3 = ctxt.x64.ctrlreg[3];

    p2m_size = info.nr_pages;
    printf("size: %lu\n", p2m_size);

    //p2m_size = get_maximum_gpfn(xc_handle, dom) + 1;
    p2m = calloc(p2m_size, sizeof(xen_pfn_t));

    if (hvm) {
        for (i = 0; i < p2m_size; i++)
            p2m[i] = i;
    }
    else {
        m2p = get_m2p_table(xc_handle);
        if (m2p == NULL)
            return;

        nr_mfns = get_maximum_gpfn(xc_handle, dom) + 1;
        mfn_list = malloc(sizeof(xen_pfn_t) * nr_mfns);

        if (xc_get_pfn_list(xc_handle, dom, mfn_list, nr_mfns) == -1) {
            printf("xc_get_pfn_list: error\n");
            goto err;
        }

        for (i = 0; i < nr_mfns; i++) {
            mfn = mfn_list[i];
            pfn = m2p[mfn];
            p2m[pfn] = mfn;
        }

    err:
        /* free(mfn_list) */ ;
    }

    err = malloc(p2m_size * sizeof(int));

    mem_base = xc_map_foreign_bulk(xc_handle, dom, PROT_READ, p2m, err,
                                   p2m_size);
    if (mem_base == NULL)
        printf("cannot map the entire memory\n");
        
    //    free(p2m);
    free(err);
}

static int ffs8(uint8_t x)
{
    int i;

    for (i = 0; i < 8; i++)
        if (x & (1u << i))
            return i + 1;

    return 0;
}

static int ffs16(uint16_t x)
{
    uint8_t h = x >> 8, l = x;
    return l ? ffs8(l) : h ? ffs8(h) + 8 : 0;
}

static int ffs32(uint32_t x)
{
    uint16_t h = x >> 16, l = x;
    return l ? ffs16(l) : h ? ffs16(h) + 16 : 0;
}

static int ffs64(uint64_t x)
{
    uint32_t h = x >> 32, l = x;
    return l ? ffs32(l) : h ? ffs32(h) + 32 : 0;
}

static unsigned long machine_to_physical(unsigned long ma)
{
    unsigned long mfn, pfn;
    unsigned long pa;

    mfn = ma >> PAGE_SHIFT;
    pfn = m2p[mfn];
    pa = pfn << PAGE_SHIFT;
    pa += ma & (PAGE_SIZE - 1);

    return pa;
}

#define PTE_PSE 0x80

static void *virtual_to_local(void *vaddr, uint64_t cr3, void *base, int hvm)
{
    uint64_t va, pa, mask, pte = 0;
    int level;
    void *laddr;

    pa = cr3 & ~0xfffull;
    va = (uint64_t)vaddr & 0x0000ffffffffffffull;
    mask = 0x0000ff8000000000ull;

    for (level = 4; level > 0; level--) {
        pa += ((va & mask) >> (ffs64(mask) - 1)) * 8;

        if (!hvm)
            pa = machine_to_physical(pa);
        laddr = base + pa;

        memcpy(&pte, laddr, 8);
        if (!(pte & 1)) 
            return NULL;

        pa = pte & 0x000ffffffffff000ull;
        if (level == 2 && (pte & PTE_PSE)) {
            mask = ((mask ^ ~-mask) >> 1);
            pa = ((pa & ~mask) | (va & mask));
            break;
        }
        mask >>= 9;
    }

    if (level != 2)
        pa += va & 0xfffull;

    if (!hvm)
        pa = machine_to_physical(pa);

    return base + pa;
}

void *addr_trans(XCH xc_handle, int dom, int cpu, uint32_t flag,
                 unsigned long addr, unsigned int size)
{
    void *laddr;

    if (mem_base == NULL) {
        addr_init(xc_handle, dom, cpu);
        printf("mem_base: %p, cr3: %lx, hvm: %d\n", mem_base, cr3, hvm);
    }

    laddr = virtual_to_local((void *)addr, cr3, mem_base, hvm);

    return laddr;
}
#endif /* MAP_ALL */
