#include "linux_types.h"
#include "linux_consts.h"
#include "array.h"
#include "addr.h"
#include "sched.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define CAP_FOR_EACH_U32(__capi)  \
	for (__capi = 0; __capi < _KERNEL_CAPABILITY_U32S; ++__capi)

#define __ALIGN_MASK(x,mask)	(((x)+(mask))&~(mask))
#define ALIGN(x,a)		__ALIGN_MASK(x,(typeof(x))(a)-1)
#define PAGE_ALIGN(addr) ALIGN(addr, PAGE_SIZE)
// asm/signal.h
#define _NSIG		64

#ifdef __i386__
# define _NSIG_BPW	32
#else
# define _NSIG_BPW	64
#endif

#define _NSIG_WORDS	(_NSIG / _NSIG_BPW)
#define CHUNKSZ				32
# define BITS_PER_LONG 64
#define smp_read_barrier_depends()	do { } while (0)
#define ACCESS_ONCE(x) (*(volatile typeof(x) *)&(x))
#define rcu_dereference(p)     ({ \
				typeof(p) _________p1 = ACCESS_ONCE(p); \
				smp_read_barrier_depends(); \
				(_________p1); \
				})
#ifdef __CHECKER__
#define __force	__attribute__((force))
#else
#define __force
#endif

#define BITOP_WORD(nr)		((nr) / BITS_PER_LONG)
// autoconf.h
#define CONFIG_NR_CPUS 64
// threads.h
#define NR_CPUS		CONFIG_NR_CPUS

#define CONFIG_NODES_SHIFT 9
#define MAX_NUMNODES    (1 << CONFIG_NODES_SHIFT)

#define CPUS 1
#define MEMS 2

// asm-generic/signal.h
#define SIG_IGN	((__force __sighandler_t)1)	
#define SIG_DFL	((__force __sighandler_t)0)

#define RLIMIT_SIGPENDING	11
#define RLIMIT_RSS		5
// cputime.h
#define cputime_add(__a, __b)		((__a) +  (__b))
#define cputime_zero			(0UL)
#define cputime_to_clock_t(__ct)	jiffies_to_clock_t(__ct)
#define clock_t_to_cputime(__x)		clock_t_to_jiffies(__x)

#define typecheck(type,x) \
({	type __dummy; \
	typeof(x) __dummy2; \
	(void)(&__dummy == &__dummy2); \
	1; \
})

#define MINORBITS	20
#define MINORMASK	((1U << MINORBITS) - 1)

#define MAJOR(dev)	((unsigned int) ((dev) >> MINORBITS))
#define MINOR(dev)	((unsigned int) ((dev) & MINORMASK))
#define MKDEV(ma,mi)	(((ma) << MINORBITS) | (mi))

#define likely(x)	__builtin_expect(!!(x), 1)
#define unlikely(x)	__builtin_expect(!!(x), 0)

// kernel config & param.h
#define CONFIG_HZ       1000
#define HZ		CONFIG_HZ
#define USER_HZ	        100

// jiffies.h
#define SH_DIV(NOM,DEN,LSH) (   (((NOM) / (DEN)) << (LSH))              \
                             + ((((NOM) % (DEN)) << (LSH)) + (DEN) / 2) / (DEN))
#define PIT_TICK_RATE 1193182
#define CLOCK_TICK_RATE	PIT_TICK_RATE
#define LATCH  ((CLOCK_TICK_RATE + HZ/2) / HZ)
#define ACTHZ (SH_DIV (CLOCK_TICK_RATE, LATCH, 8))
#define TICK_NSEC (SH_DIV (1000000UL * 1000, ACTHZ, 8))
// time.h
#define NSEC_PER_SEC	1000000000L
// sched.h
#define NICE_TO_PRIO(nice)	(MAX_RT_PRIO + (nice) + 20)
#define PRIO_TO_NICE(prio)	((prio) - MAX_RT_PRIO - 20)

# define do_div(n,base) ({					\
	uint32_t __base = (base);				\
	uint32_t __rem;						\
	__rem = ((uint64_t)(n)) % __base;			\
	(n) = ((uint64_t)(n)) / __base;				\
	__rem;							\
 })

#define MAX_DATA 256

/*
#define __acquires(x)	__attribute__((context(x,0,1)))
#define __releases(x)	__attribute__((context(x,1,0)))
#define __lockfunc __attribute__((section(".spinlock.text")))
unsigned long __lockfunc _spin_lock_irqsave(spinlock_t *lock)
                                                          __acquires(lock);

void __lockfunc _spin_unlock_irqrestore(spinlock_t *lock, unsigned long flags)
							__releases(lock);

#define spin_lock_irqsave(lock, flags)			\
	do {						\
		typecheck(unsigned long, flags);	\
		flags = _spin_lock_irqsave(lock);	\
	} while (0)

#define spin_unlock_irqrestore(lock, flags)		\
	do {						\
		typecheck(unsigned long, flags);	\
		_spin_unlock_irqrestore(lock, flags);	\
	} while (0)
*/
// kernel.h
#define min(x, y) ({				\
	typeof(x) _min1 = (x);			\
	typeof(y) _min2 = (y);			\
	(void) (&_min1 == &_min2);		\
	_min1 < _min2 ? _min1 : _min2; })

#define max(x, y) ({				\
	typeof(x) _max1 = (x);			\
	typeof(y) _max2 = (y);			\
	(void) (&_max1 == &_max2);		\
	_max1 > _max2 ? _max1 : _max2; })

unsigned long get_mm_rss(struct mm_struct *mm)
{
  return mm->_file_rss.counter + mm->_anon_rss.counter;
}

struct task_struct *next_thread(struct task_struct *addr, struct sched_struct *handle)
{
    struct list_head **usr_thread_next = HREF(addr, thread_group.next);

    return (struct task_struct *)((unsigned long)(*usr_thread_next) - (offsetof(struct task_struct, thread_group.next)));
}

static inline u32 new_encode_dev(dev_t dev)
{
  unsigned major = MAJOR(dev);
  unsigned minor = MINOR(dev);
  return (minor & 0xff) | (major << 8) | ((minor & ~0xff) << 12);
}

static inline u64 div_u64_rem(u64 dividend, u32 divisor, u32 *remainder)
{
	*remainder = dividend % divisor;
	return dividend / divisor;
}

static inline u64 div_u64(u64 dividend, u32 divisor)
{
	u32 remainder;
	return div_u64_rem(dividend, divisor, &remainder);
}

u64 nsec_to_clock_t(u64 x)
{
#if (NSEC_PER_SEC % USER_HZ) == 0
	return div_u64(x, NSEC_PER_SEC / USER_HZ);
#elif (USER_HZ % 512) == 0
	return div_u64(x * USER_HZ / 512, NSEC_PER_SEC / 512);
#else
     	return div_u64(x * 9, (9ull * NSEC_PER_SEC + (USER_HZ / 2)) / USER_HZ);
#endif
}

unsigned long clock_t_to_jiffies(unsigned long x)
{
#if (HZ % USER_HZ)==0
	if (x >= ~0UL / (HZ / USER_HZ))
		return ~0UL;
	return x * (HZ / USER_HZ);
#else
	/* Don't worry about loss of precision here .. */
	if (x >= ~0UL / HZ * USER_HZ)
		return ~0UL;

	/* .. but do try to contain it here */
	return div_u64((u64)x * HZ, USER_HZ);
#endif
}

clock_t jiffies_to_clock_t(long x)
{
#if (TICK_NSEC % (NSEC_PER_SEC / USER_HZ)) == 0
# if HZ < USER_HZ
	return x * (USER_HZ / HZ);
# else
	return x / (HZ / USER_HZ);
# endif
#else
	return div_u64((u64)x * TICK_NSEC, NSEC_PER_SEC / USER_HZ);
#endif
}

cputime_t task_utime(struct task_struct *p)
{
  clock_t utime = cputime_to_clock_t(p->utime),
    total = utime + cputime_to_clock_t(p->stime);
  u64 temp;
  temp = (u64)nsec_to_clock_t(p->se.sum_exec_runtime);
  
  if (total) {
    temp *= utime;
    do_div(temp, total);
  }
  utime = (clock_t)temp;
  return max(p->prev_utime, clock_t_to_cputime(utime));
}

/* CONFIG_VIRT_CPU_ACCOUNTING
cputime_t task_utime(struct task_struct *p,
		     int xc_handle, int dom, int cpu)
{
  cputime_t *ptr_utime = (cputime_t *)((unsigned long)p + offsetof(struct task_struct, utime));
  cputime_t *user_ptr_utime = addr_trans(xc_handle, dom, cpu, PROT_READ, (uint64_t)ptr_utime);
  return *user_ptr_utime;
}
*/

cputime_t task_stime(struct task_struct *p)
{
  clock_t stime;
  cputime_t temp;

  stime = nsec_to_clock_t(p->se.sum_exec_runtime) -
    cputime_to_clock_t(task_utime(p));

  if (stime >= 0)
    temp = max(p->prev_stime, clock_t_to_cputime(stime));
  else
    temp = p->prev_stime;

  return temp;
}

/* CONFIG_VIRT_CPU_ACCOUNTING
cputime_t task_stime(struct task_struct *p,
		     int xc_handle, int dom, int cpu)
{
  cputime_t *ptr_stime = (cputime_t *)((unsigned long)p + offsetof(struct task_struct, stime));
  cputime_t *user_ptr_stime = addr_trans(xc_handle, dom, cpu, PROT_READ, (uint64_t)ptr_stime);
  return *user_ptr_stime;
}
*/
inline cputime_t task_gtime(struct task_struct *p)
{
  return p->gtime;
}


dev_t tty_devnum(struct tty_struct *tty, struct sched_struct *handle)
{
  struct tty_driver **user_ptr_driver = HREF(tty, driver);
  int *user_ptr_major = HREF(*user_ptr_driver, major);
  int *user_ptr_minor_start = HREF(*user_ptr_driver, minor_start);
  int *user_ptr_index = HREF(tty, index);

  return MKDEV(*user_ptr_major, *user_ptr_minor_start) + *user_ptr_index;
}

void task_mem(char *sp, struct mm_struct *mm)
{
  char str[MAX_DATA];
  
  unsigned long data, text, lib;
  unsigned long hiwater_vm, total_vm, hiwater_rss, total_rss;
  
  hiwater_vm = total_vm = mm->total_vm;
  if (hiwater_vm < mm->hiwater_vm)
    hiwater_vm = mm->hiwater_vm;
  hiwater_rss = total_rss = get_mm_rss(mm);
  if (hiwater_rss < mm->hiwater_rss)
    hiwater_rss = mm->hiwater_rss;
  
  data = mm->total_vm - mm->shared_vm - mm->stack_vm;
  text = (PAGE_ALIGN(mm->end_code) - (mm->start_code & PAGE_MASK)) >> 10;
  lib = (mm->exec_vm << (PAGE_SHIFT-10)) - text;
  sprintf(str,
	  "VmPeak:\t%8lu kB\n"
	  "VmSize:\t%8lu kB\n"
	  "VmLck:\t%8lu kB\n"
	  "VmHWM:\t%8lu kB\n"
	  "VmRSS:\t%8lu kB\n"
	  "VmData:\t%8lu kB\n"
	  "VmStk:\t%8lu kB\n"
	  "VmExe:\t%8lu kB\n"
	  "VmLib:\t%8lu kB\n"
	  "VmPTE:\t%8lu kB\n",
	  hiwater_vm << (PAGE_SHIFT-10),
	  (total_vm - mm->reserved_vm) << (PAGE_SHIFT-10),
	  mm->locked_vm << (PAGE_SHIFT-10),
	  hiwater_rss << (PAGE_SHIFT-10),
	  total_rss << (PAGE_SHIFT-10),
	  data << (PAGE_SHIFT-10),
	  mm->stack_vm << (PAGE_SHIFT-10), text, lib,
	  (PTRS_PER_PTE*sizeof(pte_t)*mm->nr_ptes) >> 10);
  
  strcat(sp, str);
}

static inline int task_ptrace(struct task_struct *task)
{
  return task->ptrace;
}

static inline struct task_struct *tracehook_tracer_task(struct task_struct *tsk)
{
  
  if (task_ptrace(tsk) & PT_PTRACED){
    return tsk->parent;
  }
  return NULL;
}

struct pid *task_pid(struct task_struct *addr, struct sched_struct *handle)
{
    struct pid **usr_pid = HREF(addr, pids[PIDTYPE_PID].pid);
    struct pid *tmp = *usr_pid;
    UNMAP(usr_pid);
    return tmp;
}

struct pid *task_tgid(struct task_struct *task, struct sched_struct *handle)
{
    //  struct pid *ptr_pid = (struct pid *)((unsigned long)task->group_leader + offsetof(struct task_struct, pids[PIDTYPE_PID]) + offsetof(struct pid_link, pid));
  struct pid_link *pids = GREF(task->group_leader, pids[PIDTYPE_PID]);
  struct pid **user_ptr_pid = HREF(pids, pid);

  return *user_ptr_pid;
}

pid_t pid_nr_ns(struct pid *pid, struct pid_namespace *ns, struct sched_struct *handle)
{
  pid_t nr = 0;

  unsigned int *user_ptr_ns_level = HREF(ns, level);
  unsigned int *user_ptr_pid_level = HREF(pid, level);

  if (pid && *user_ptr_ns_level <= *user_ptr_pid_level) {
    struct upid *upid = GREF(pid, numbers[*user_ptr_ns_level]);
    struct pid_namespace **user_ptr_pid_numbers = HREF(upid, ns);

    if (*user_ptr_pid_numbers == ns) {
      int *user_ptr_upid_nr = HREF(upid, nr);
      nr = *user_ptr_upid_nr;
    }
  }
  return nr;
}

pid_t task_pid_nr_ns(struct task_struct *addr, struct pid_namespace *ns, struct sched_struct *handle)
{
  return pid_nr_ns(task_pid(addr, handle), ns, handle);
}


pid_t task_tgid_nr_ns(struct task_struct *tsk, struct pid_namespace *ns, struct sched_struct *handle)
{
  return pid_nr_ns(task_tgid(tsk, handle), ns, handle);
}

pid_t pid_vnr(struct task_struct *addr, struct pid_namespace *ns, struct sched_struct *handle)
{
	return pid_nr_ns(task_pid(addr, handle), ns, handle);
}

inline int pid_alive(struct task_struct *task)
{
  return task->pids[PIDTYPE_PID].pid != NULL;
}

struct mm_struct *get_task_mm(struct task_struct *task, struct sched_struct *handle)
{
  struct mm_struct *mm = task->mm;

  //	task_lock(task);
  if (mm) {
    if (task->flags & PF_KTHREAD)
      mm = NULL;
    else
      {
          mm = ADDRX(mm);
        struct mmu_notifier_mm **user_mm_mmu = HREF(task->mm, mmu_notifier_mm);
       
	/*
	volatile struct mm_struct * ptr_mm = (volatile struct mm_struct*)((unsigned long)task + offsetof(struct task_struct, mm));
	volatile struct mm_struct **user_ptr_mm = addr_trans(xc_handle, dom, cpu, PROT_READ, (uint64_t)ptr_mm);
	volatile int *ptr_counter = (volatile int *)((unsigned long)*user_ptr_mm + offsetof(struct mm_struct,mm_users) + offsetof(atomic_t, counter));
	volatile int *user_ptr_counter = addr_trans(xc_handle, dom, cpu, PROT_READ, (uint64_t)ptr_counter);
	      	asm volatile(LOCK_PREFIX "incl %0"
		     : "=m" (user_ptr_counter)
		     : "m" (user_ptr_counter));
	*/
      }
  }
  //	task_unlock(task);
  return mm;
}

inline void task_name(char *sp, struct task_struct *p)
{
  char str[MAX_DATA];

  sprintf(str, "Name:\t%s\n", p->comm);
  strcat(sp, str);
}

static const char *task_state_array[] = {
  "R (running)",	/*  0 */
  "S (sleeping)",	/*  1 */
  "D (disk sleep)",	/*  2 */
  "T (stopped)",	/*  4 */
  "T (tracing stop)",	/*  8 */
  "Z (zombie)",		/* 16 */
  "X (dead)"		/* 32 */
};

const char *get_task_state(struct task_struct *tsk)
{
  unsigned int state = (tsk->state & TASK_REPORT) | tsk->exit_state;
  const char **p = &task_state_array[0];

  while(state){
    p++;
    state >>= 1;
  }

  return *p;
}

inline void task_state(char *sp, struct pid_namespace *ns, struct pid *pid, struct task_struct *task, struct sched_struct *handle)
{
  char str[MAX_DATA];
  struct group_info *group_info;
  int g;
  struct fdtable *fdt = NULL;
  unsigned int max_fds = 0;
  pid_t ppid, tpid;
  struct task_struct *user_real_parent_map = ADDRX(task->real_parent);
  int *user_latency_count = HREF(task->real_parent, latency_record_count);

  ppid = pid_alive(task) ?
      task_tgid_nr_ns(user_real_parent_map, ns, handle) : 0;
  
  tpid = 0;
  
  if (pid_alive(task)) {
    struct task_struct *tracer = tracehook_tracer_task(task);
    if (tracer)
      tpid = task_pid_nr_ns(tracer, ns, handle);
  }

  sprintf(str,
	  "State:\t%s\n"
	  "Tgid:\t%d\n"
	  "Pid:\t%d\n"
	  "PPid:\t%d\n"
	  "TracerPid:\t%d\n"
	  "Uid:\t%d\t%d\t%d\t%d\n"
	  "Gid:\t%d\t%d\t%d\t%d\n",
	  get_task_state(task),
	  task_tgid_nr_ns(task, ns, handle),
	  pid_nr_ns(pid, ns, handle),
	  ppid, tpid,
	  task->uid, task->euid, task->suid, task->fsuid,
	  task->gid, task->egid, task->sgid, task->fsgid);
  strcat(sp, str);

  if (task->files){
    struct fdtable **user_ptr_files_fdt = HREF(task->files, fdt);
    fdt = *user_ptr_files_fdt;

    unsigned int *user_ptr_max_fds = HREF(fdt, max_fds);
    max_fds = *user_ptr_max_fds;
  }

  sprintf(str,
	  "FDSize:\t%d\n"
	  "Groups:\t",
	  fdt ? max_fds : 0);
  strcat(sp, str);

  group_info = task->group_info;
  //get_group_info(groupinfo)

  int *user_ptr_ngroups = HREF(group_info, ngroups);

  for (g = 0; g < min(*user_ptr_ngroups, NGROUPS_SMALL); g++){
    gid_t **user_ptr_blocks = HREF(group_info, blocks[g/NGROUPS_PER_BLOCK]);

    gid_t *ptr_blocks_second = (gid_t *)((unsigned long)*user_ptr_blocks + sizeof(gid_t)*g%NGROUPS_PER_BLOCK);
    gid_t *user_ptr_blocks_second = ADDRX(ptr_blocks_second);

    sprintf(str, "%d ", *user_ptr_blocks_second);
    strcat(sp, str);
  }

  //put_group_info(group_info)
  strcat(sp, "\n");
}

static inline int sigismember(kern_sigset_t *set, int _sig)
{
  unsigned long sig = _sig - 1;
  
  if (_NSIG_WORDS == 1)
    return 1 & (set->sig[0] >> sig);
  else
    return 1 & (set->sig[sig / _NSIG_BPW] >> (sig % _NSIG_BPW));
}

void render_sigset_t(char *sp, const char *header, kern_sigset_t *set)
{
  int i;
  char str[MAX_DATA];

  sprintf(str, "%s", header);
  strcat(sp, str);

  i = _NSIG;
  do {
    int x = 0;
    
    i -= 4;
    if (sigismember(set, i+1)) x |= 1;
    if (sigismember(set, i+2)) x |= 2;
    if (sigismember(set, i+3)) x |= 4;
    if (sigismember(set, i+4)) x |= 8;
    sprintf(str, "%x", x);
    strcat(sp, str);
  } while (i >= 4);
  
  strcat(sp, "\n");
}

static inline void sigemptyset(kern_sigset_t *set)
{
  switch (_NSIG_WORDS) {
  default:
    memset(set, 0, sizeof(kern_sigset_t));
    break;
  case 2: set->sig[1] = 0;
  case 1:	set->sig[0] = 0;
    break;
  }
}

static inline void sigaddset(kern_sigset_t *set, int _sig)
{
  unsigned long sig = _sig - 1;
  if (_NSIG_WORDS == 1)
    set->sig[0] |= 1UL << sig;
  else
    set->sig[sig / _NSIG_BPW] |= 1UL << (sig % _NSIG_BPW);
}

void collect_sigign_sigcatch(struct task_struct *p, kern_sigset_t *ign, kern_sigset_t *catch, struct sched_struct *handle)
{
  int i = 0;

  for (i = 1; i <= _NSIG; ++i) {
    __sighandler_t *ptr_handler= (__sighandler_t *)((unsigned long)p->sighand + offsetof(struct sighand_struct, action[i-1]) + 
							offsetof(struct k_sigaction, sa) + offsetof(struct kern_sigaction, sa_handler));
    __sighandler_t *user_ptr_handler = ADDRX(ptr_handler);
    if (*user_ptr_handler == SIG_IGN)
      sigaddset(ign, i);
    else if (*user_ptr_handler != SIG_DFL)
      sigaddset(catch, i);
  }
}

struct sighand_struct *lock_task_sighand(struct task_struct *tsk, unsigned long *flags, struct sched_struct *handle)
{
  struct sighand_struct *sighand;
  
  //rcu_read_lock();
    for (;;) {
      sighand = rcu_dereference(tsk->sighand);
      if (unlikely(sighand == NULL))
	break;

      spinlock_t *user_ptr_siglock = HREF(tsk->sighand, siglock);

      // spin_lock_irqsave(user_ptr_siglock, *flags);
      if (likely(sighand == tsk->sighand))
	break;
      // spin_unlock_irqrestore(&sighand->siglock, *flags);
    }
  //rcu_read_unlock();

  return sighand;
}


inline void task_sig(char *sp, struct task_struct *p, struct sched_struct *handle)
{
  char str[MAX_DATA];
  unsigned long flags;
  kern_sigset_t pending, shpending, blocked, ignored, caught;
  int num_threads = 0;
  unsigned long qsize = 0;
  unsigned long qlim = 0;

  sigemptyset(&pending);
  sigemptyset(&shpending);
  sigemptyset(&blocked);
  sigemptyset(&ignored);
  sigemptyset(&caught);

  //rcu_read_lock();
  if (lock_task_sighand(p, &flags, handle)) {
    pending = p->pending.signal;

    kern_sigset_t *ptr_sharedpending_signal = (kern_sigset_t *)((unsigned long)p->signal + offsetof(struct signal_struct, shared_pending) + offsetof(struct sigpending, signal));
    kern_sigset_t *user_ptr_sharedpending_signal = ADDRX(ptr_sharedpending_signal);
    shpending = *user_ptr_sharedpending_signal;

    blocked = p->blocked;
    collect_sigign_sigcatch(p, &ignored, &caught, handle);

    int *ptr_signal_count = (int *)((unsigned long)p->signal + offsetof(struct signal_struct, count) + offsetof(atomic_t, counter));
    int *user_ptr_signal_count = ADDRX(ptr_signal_count);
    num_threads = *user_ptr_signal_count;

    int *ptr_user_count = (int *)((unsigned long)p->user + offsetof(struct user_struct, sigpending) + offsetof(atomic_t, counter));
    int *user_ptr_user_count = ADDRX(ptr_user_count);
    qsize = *user_ptr_user_count;

    long unsigned int *ptr_rlim_cur = (long unsigned int *)((unsigned long)p->signal + offsetof(struct signal_struct, rlim[RLIMIT_SIGPENDING]) + offsetof(struct kern_rlimit, rlim_cur));
    long unsigned int *user_ptr_rlim_cur = ADDRX(ptr_rlim_cur);
    qlim = *user_ptr_rlim_cur;
    
    //unlock_task_sighand(p, &flags);
  }
  //rcu_read_unlock();
 
 sprintf(str, "Threads:\t%d\n", num_threads);
 strcat(sp, str);
 sprintf(str, "SigQ:\t%lu/%lu\n", qsize, qlim);
 strcat(sp, str);

 render_sigset_t(sp, "SigPnd:\t", &pending);
 render_sigset_t(sp, "ShdPnd:\t", &shpending);
 render_sigset_t(sp, "SigBlk:\t", &blocked);
 render_sigset_t(sp, "SigIgn:\t", &ignored);
 render_sigset_t(sp, "SigCgt:\t", &caught);
}

inline void task_cap(char *sp, struct task_struct *p)
{
  unsigned __capi;
  char str[MAX_DATA];
  
  sprintf(str, "CapInh:\t");
  strcat(sp, str);
  CAP_FOR_EACH_U32(__capi) {
    sprintf(str, "%08x", p->cap_inheritable.cap[(_KERNEL_CAPABILITY_U32S-1) - __capi]);
    strcat(sp, str);
  }
  strcat(sp, "\n");

  sprintf(str, "CapPrm:\t");
  strcat(sp, str);
  CAP_FOR_EACH_U32(__capi) {
    sprintf(str, "%08x", p->cap_permitted.cap[(_KERNEL_CAPABILITY_U32S-1) - __capi]);
    strcat(sp, str);
  }
  strcat(sp, "\n");

  sprintf(str, "CapEff:\t");
  strcat(sp, str);
    CAP_FOR_EACH_U32(__capi) {
    sprintf(str, "%08x", p->cap_effective.cap[(_KERNEL_CAPABILITY_U32S-1) - __capi]);
    strcat(sp, str);
  }
  strcat(sp, "\n");

  sprintf(str, "CapBnd:\t");
  strcat(sp, str);
  CAP_FOR_EACH_U32(__capi) {
    sprintf(str, "%08x", p->cap_bset.cap[(_KERNEL_CAPABILITY_U32S-1) - __capi]);
    strcat(sp, str);
  }
  strcat(sp, "\n");
}

inline void task_context_switch_counts(char *sp, struct task_struct *p)
{
  char str[MAX_DATA];
  sprintf(str, "voluntary_ctxt_switches:\t%lu\n"
	  "nonvoluntary_ctxt_switches:\t%lu\n",
	  p->nvcsw,
	  p->nivcsw);
  strcat(sp, str);
}

static inline struct pid *task_session(struct task_struct *task, struct sched_struct *handle)
{
  struct pid *ptr_pid = (struct pid *)((unsigned long)task->group_leader + offsetof(struct task_struct, pids[PIDTYPE_SID]) + offsetof(struct pid_link, pid));
  struct pid **user_ptr_pid = ADDRX(ptr_pid);
  return *user_ptr_pid;
}

pid_t task_session_nr_ns(struct task_struct *tsk, struct pid_namespace *ns, struct sched_struct *handle)
{
  return pid_nr_ns(task_session(tsk, handle), ns, handle);
}

void bitmap_scnprintf(char *sp, struct task_struct *task, int nmaskbits, int flag)
{
  char str[MAX_DATA];
  int i, word, bit, len = 0;
  unsigned long val;
  const char *sep = "";
  int chunksz;
  u32 chunkmask;
  
  chunksz = nmaskbits & (CHUNKSZ - 1);
  if (chunksz == 0)
    chunksz = CHUNKSZ;
  
  i = ALIGN(nmaskbits, CHUNKSZ) - CHUNKSZ;
  for (; i >= 0; i -= CHUNKSZ) {
    chunkmask = ((1ULL << chunksz) - 1);
    word = i / BITS_PER_LONG;
    bit = i % BITS_PER_LONG; 
    if(flag == CPUS){
      val = (task->cpus_allowed.bits[word] >> bit) & chunkmask;
    }else if(flag == MEMS){
      val = (task->mems_allowed.bits[word] >> bit) & chunkmask;
    }
    sprintf(str, "%s%0*lx", sep, (chunksz+3)/4, val);
    strcat(sp, str);
    
    chunksz = CHUNKSZ;
    sep = ",";
  }
}

static inline unsigned long __ffs(unsigned long word)
{
	asm("bsf %1,%0"
		: "=r" (word)
		: "rm" (word));
	return word;
}

unsigned long find_first_bit(struct task_struct *task, unsigned long size, int flag)
{
  unsigned long p;
  unsigned long result = 0;
  unsigned long tmp;
  int count = 0;

  while (size & ~(BITS_PER_LONG-1)) {
    if(flag == CPUS){
      tmp = task->cpus_allowed.bits[count];
    }else if (flag == MEMS){
      tmp = task->mems_allowed.bits[count];
    }
    count++;
    if ((tmp)){
      goto found;
    }
    result += BITS_PER_LONG;
    size -= BITS_PER_LONG;
  }
  if (!size)
    return result;
  if(flag == CPUS){
    p = task->cpus_allowed.bits[count];
  }else if (flag == MEMS){
    p = task->mems_allowed.bits[count];
  }
  
  tmp = (p) & (~0UL >> (BITS_PER_LONG - size));
  if (tmp == 0UL)		/* Are any bits set? */
    return result + size;	/* Nope. */
 found:
  return result + __ffs(tmp);
}

unsigned long find_next_bit(struct task_struct *task, unsigned long size, unsigned long offset, int flag)
{
  const unsigned long *p;
  unsigned long result = offset & ~(BITS_PER_LONG-1);
  unsigned long tmp;
  int count = BITOP_WORD(offset);
  if(flag == CPUS){
    tmp = task->cpus_allowed.bits[count];
  } else if (flag == MEMS){
    tmp = task->mems_allowed.bits[count];
  }
  count++;
  if (offset >= size)
    return size;
  size -= result;
  offset %= BITS_PER_LONG;
  if (offset) {
    tmp &= (~0UL << offset);
    if (size < BITS_PER_LONG)
      goto found_first;
    if (tmp)
      goto found_middle;
    size -= BITS_PER_LONG;
    result += BITS_PER_LONG;
  }
  while (size & ~(BITS_PER_LONG-1)) {
    if (tmp){
      if(flag == CPUS){
	tmp = task->cpus_allowed.bits[count];
      } else if (flag == MEMS){
	tmp = task->mems_allowed.bits[count];
      }
      goto found_middle;
    }
    result += BITS_PER_LONG;
    size -= BITS_PER_LONG;
  }
  if (!size)
    return result;
  if(flag == CPUS){
    tmp = task->cpus_allowed.bits[count-1];
  } else if (flag == MEMS){
    tmp = task->mems_allowed.bits[count-1];
  }
  
found_first:
  tmp &= (~0UL >> (BITS_PER_LONG - size));
  if (tmp == 0UL)		/* Are any bits set? */
    return result + size;	/* Nope. */
found_middle:
  return result + __ffs(tmp);
}

static inline int bscnl_emit(char *sp, int rbot, int rtop, int len)
{
  char str[MAX_DATA];
  if (len > 0){
    len += sprintf(str, ",");
    strcat(sp, str);
  }
  if (rbot == rtop){
    len += sprintf(str, "%d", rbot);
    strcat(sp, str);
  } else {
    len += sprintf(str, "%d-%d", rbot, rtop);
    strcat(sp, str);
  }
  return len;
}

void bitmap_scnlistprintf(char *sp, struct task_struct *task, int nmaskbits, int flag)
{
  int len = 0;
  /* current bit is 'cur', most recently seen range is [rbot, rtop] */
  int cur, rbot, rtop;

  /*
  if (buflen == 0)
    return 0;
  */
  rbot = cur = find_first_bit(task, nmaskbits, flag);
  while (cur < nmaskbits) {
    rtop = cur;
    cur = find_next_bit(task, nmaskbits, cur+1, flag);
    if (cur >= nmaskbits || cur > rtop + 1) {
      len = bscnl_emit(sp, rbot, rtop, len);
      rbot = cur;
    }
  }
}

void cpuset_task_status_allowed(char *sp, struct task_struct *task)
{
  char str[MAX_DATA];
  sprintf(str, "Cpus_allowed:\t");
  strcat(sp, str);
  bitmap_scnprintf(sp, task, NR_CPUS, CPUS);
  strcat(sp, "\n");
  
  sprintf(str, "Cpus_allowed_list:\t");
  strcat(sp, str);
  bitmap_scnlistprintf(sp, task, NR_CPUS, CPUS);
  strcat(sp, "\n");
  
  sprintf(str, "Mems_allowed:\t");
  strcat(sp, str);
  bitmap_scnprintf(sp, task, MAX_NUMNODES, MEMS);
  strcat(sp, "\n");
  
  sprintf(str, "Mems_allowed_list:\t");
  strcat(sp, str);
  bitmap_scnlistprintf(sp, task, NR_CPUS, MEMS);
  strcat(sp, "\n");
}


int proc_pid_status(char *sp, struct task_struct *addr, struct sched_struct *handle)
{
    memset(sp, 0, sizeof(char));
    struct pid *pid = task_pid(addr, handle);
    struct pid_namespace **usr_ns = HREF(pid, numbers[0].ns);
    struct task_struct *task = get_task_struct(addr, handle);

    struct mm_struct *mm = get_task_mm(task, handle);
    task_name(sp, task);

  task_state(sp, *usr_ns, pid, task, handle);
  if (mm) {
    task_mem(sp, mm);
    //mmput(mm);
  }
  task_sig(sp, task, handle);
  task_cap(sp, task);
  cpuset_task_status_allowed(sp, task);

  /*
#if defined(CONFIG_S390)
  task_show_regs(m, task);
#endif
  */
  task_context_switch_counts(sp, task);

  UNMAP(usr_ns);
  UNMAP(task);
  return 0;
}

int task_prio(const struct task_struct *p)
{
  return p->prio - MAX_RT_PRIO;
}

int task_nice(const struct task_struct *p)
{
  return PRIO_TO_NICE(p->static_prio);
}

static inline unsigned int task_cpu(const struct task_struct *p, struct sched_struct *handle)
{
    //  __u32 *ptr_cpu = (__u32 *)((unsigned long)p->stack + offsetof(struct thread_info, cpu));
  __u32 *user_ptr_cpu = HREF((struct thread_info *)p->stack, cpu);

  return *user_ptr_cpu;
}

u64 __delayacct_blkio_ticks(struct task_struct *tsk, struct sched_struct *handle)
{
  u64 ret;
  unsigned long flags;

  u64 *user_ptr_blkio_delay = HREF(tsk->delays, blkio_delay);
  u64 *user_ptr_swapin_delay = HREF(tsk->delays, swapin_delay);
  //spin_lock_irqsave(&tsk->delays->lock, flags);
  ret = nsec_to_clock_t(*user_ptr_blkio_delay +
			*user_ptr_swapin_delay);
  //spin_unlock_irqrestore(&tsk->delays->lock, flags);
  return ret;
}

static inline u64 delayacct_blkio_ticks(struct task_struct *tsk, struct sched_struct *handle)
{
  if (tsk->delays)
    return __delayacct_blkio_ticks(tsk, handle);
  return 0;
}

static inline struct pid *task_pgrp(struct task_struct *task, struct sched_struct *handle)
{
  struct pid *ptr_pid = (struct pid *)((unsigned long)task->group_leader + offsetof(struct task_struct, pids[PIDTYPE_PGID]) + offsetof(struct pid_link, pid));
  struct pid **user_ptr_pid = ADDRX(ptr_pid);

  return *user_ptr_pid;
}

pid_t task_pgrp_nr_ns(struct task_struct *tsk, struct pid_namespace *ns, struct sched_struct *handle)
{
  return pid_nr_ns(task_pgrp(tsk, handle), ns, handle);
}

unsigned long task_vsize(struct mm_struct *mm)
{
  return PAGE_SIZE * (mm->total_vm);
}

int do_task_stat(char *sp, struct task_struct *addr, int whole, struct sched_struct *handle)
{
	struct task_struct *task = get_task_struct(addr, handle);
  unsigned long vsize, eip, esp, wchan = ~0UL;
  long priority, nice;
  int tty_pgrp = -1, tty_nr = 0;
  kern_sigset_t sigign, sigcatch;
  char state;
  pid_t ppid = 0, pgid = -1, sid = -1;
  int num_threads = 0;
  int permitted;
  struct mm_struct *mm;
  unsigned long long start_time;
  unsigned long cmin_flt = 0, cmaj_flt = 0;
  unsigned long  min_flt = 0,  maj_flt = 0;
  cputime_t cutime, cstime, utime, stime;
  cputime_t cgtime, gtime;
  unsigned long rsslim = 0;
  unsigned long flags;
  long unsigned int start_code = 0, end_code = 0, start_stack = 0;
	struct pid *pid = task_pid(addr, handle);
	struct pid_namespace **usr_ns = HREF(pid, numbers[0].ns);
	
  state = *get_task_state(task);
  vsize = eip = esp = 0;
  //permitted = ptrace_may_access(task, PTRACE_MODE_READ);
  mm = get_task_mm(task, handle);
  
  if (mm) {
    vsize = task_vsize(mm);
    /*
    if (permitted) {
      eip = KSTK_EIP(task);
      esp = KSTK_ESP(task);
    }
    */
  }
  sigemptyset(&sigign);
  sigemptyset(&sigcatch);
  cutime = cstime = utime = stime = cputime_zero;
  cgtime = gtime = cputime_zero;
  
  if (lock_task_sighand(task, &flags, handle)) {
      struct signal_struct *sig = ADDRX(task->signal);
      struct tty_audit_buf **user_ptr_tty_buf = HREF(task->signal, tty_audit_buf);

    if (sig->tty) {
      //struct pid *pgrp = tty_get_pgrp(*user_ptr_tty); 
      struct pid **user_ptr_pgrp = HREF(sig->tty, pgrp);
      tty_pgrp = pid_nr_ns(*user_ptr_pgrp, *usr_ns, handle);
      //put_pid(pgrp);
      tty_nr = new_encode_dev(tty_devnum(sig->tty, handle));
    }

    num_threads = sig->count.counter;
    collect_sigign_sigcatch(task, &sigign, &sigcatch, handle);
    
    cmin_flt = sig->cmin_flt;
    cmaj_flt = sig->cmaj_flt;
    cutime = sig->cutime;
    cstime = sig->cstime;
    cgtime = sig->cgtime;
    rsslim = sig->rlim[RLIMIT_RSS].rlim_cur;
    
    /* add up live thread stats at the group level */
    if (whole) {
      struct task_struct *t = addr;
      do {
	struct task_struct *tmp = get_task_struct(t, handle);
	min_flt += tmp->min_flt;
	maj_flt += tmp->maj_flt;

	utime = cputime_add(utime, task_utime(tmp));
	stime = cputime_add(stime, task_stime(tmp));
	gtime = cputime_add(gtime, task_gtime(tmp));
	t = next_thread(t, handle);
	UNMAP(tmp);
      } while (t != addr);

      min_flt += sig->min_flt;
      maj_flt += sig->maj_flt;
      utime = cputime_add(utime, sig->utime);
      stime = cputime_add(stime, sig->stime);
      gtime = cputime_add(gtime, sig->gtime);
    }
    
    sid = task_session_nr_ns(task, *usr_ns,handle); 
    struct task_struct *user_ptr_real_parent = get_task_struct(task->real_parent, handle);
    ppid = task_tgid_nr_ns(user_ptr_real_parent, *usr_ns, handle);
    pgid = task_pgrp_nr_ns(task, *usr_ns, handle);
    
    //unlock_task_sighand(task, &flags);
  }
  
  /*if (permitted && (!whole || num_threads < 2))
    wchan = get_wchan(task);*/
  
  if (!whole) {
    min_flt = task->min_flt;
    maj_flt = task->maj_flt;
    utime = task_utime(task);
    stime = task_stime(task);
    gtime = task_gtime(task);
  }
  
  priority = task_prio(task);
  nice = task_nice(task);
  
  start_time =
    (unsigned long long)task->real_start_time.tv_sec * NSEC_PER_SEC + task->real_start_time.tv_nsec;
  /* convert nsec -> ticks */
  
  start_time = nsec_to_clock_t(start_time);
  if(mm){
    start_code = mm->start_code;
    end_code = mm->end_code;
    start_stack = mm->start_stack;
  }
  sprintf(sp, "%d (%s) %c %d %d %d %d %d %u %lu \
%lu %lu %lu %lu %lu %ld %ld %ld %ld %d 0 %llu %lu %ld %lu %lu %lu %lu %lu \
%lu %lu %lu %lu %lu %lu %lu %lu %d %d %u %u %llu %lu %ld\n",
	     pid_nr_ns(pid, *usr_ns, handle),
	     task->comm,
	     state,
	     ppid,
	     pgid,
	     sid,
	     tty_nr,
	     tty_pgrp,
	     task->flags,
	     min_flt,
	     cmin_flt,
	     maj_flt,
	     cmaj_flt,
	     cputime_to_clock_t(utime),
	     cputime_to_clock_t(stime),
	     cputime_to_clock_t(cutime),
	     cputime_to_clock_t(cstime),
	     priority,
	     nice,
	     num_threads,
	     start_time,
	     vsize,
	  mm ? get_mm_rss(mm) : 0,
	     rsslim,
	     start_code,
	     end_code,
	     (permitted && mm) ? start_stack : 0,
	     esp,
	     eip,
	     task->pending.signal.sig[0] & 0x7fffffffUL,
	     task->blocked.sig[0] & 0x7fffffffUL,
	     sigign      .sig[0] & 0x7fffffffUL,
	     sigcatch    .sig[0] & 0x7fffffffUL,
	     wchan,
	     0UL,
	     0UL,
	     task->exit_signal,
	  task_cpu(task, handle),
	     task->rt_priority,
	     task->policy,
	  (unsigned long long)delayacct_blkio_ticks(task, handle),
	     cputime_to_clock_t(gtime),
	     cputime_to_clock_t(cgtime));
  
  //if (mm)
  // mmput(mm);
  UNMAP(task);
  UNMAP(usr_ns);
  return 0;
}

int proc_tid_stat(char *sp, struct task_struct *addr, struct sched_struct *handle)
{
  return do_task_stat(sp, addr, 0, handle);
}

int proc_tgid_stat(char *sp, struct task_struct *addr, struct sched_struct *handle)
{
  return do_task_stat(sp, addr, 1, handle);
}

int proc_pid_statm(char *sp, struct pid_namespace *ns, struct pid *pid, 
		   struct task_struct *task, struct sched_struct *handle)
{
	int size = 0, resident = 0, shared = 0, text = 0, lib = 0, data = 0;
	struct mm_struct *mm = get_task_mm(task, handle);
	/*
	if (mm) {
	  //size = task_statm(mm, &shared, &text, &data, &resident);
	  //	mmput(mm);
	}
	
	seq_printf(m, "%d %d %d %d %d %d %d\n",
			size, resident, shared, text, lib, data, 0);
	*/

	return 0;
}
