#ifndef __LINUX_TYPES_H
#define __LINUX_TYPES_H

#ifdef CONFIG_X86_32
#define BITS_PER_LONG 32
#else
#define BITS_PER_LONG 64
#endif

#define CONFIG_X86_L1_CACHE_SHIFT 7
#define L1_CACHE_SHIFT  (CONFIG_X86_L1_CACHE_SHIFT)
#define L1_CACHE_BYTES  (1 << L1_CACHE_SHIFT)

#ifndef SMP_CACHE_BYTES
#define SMP_CACHE_BYTES L1_CACHE_BYTES
#endif

#ifndef ____cacheline_aligned
#define ____cacheline_aligned __attribute__((__aligned__(SMP_CACHE_BYTES)))
#endif

typedef struct raw_spinlock {
    unsigned int slock;
} raw_spinlock_t;

typedef struct {
    raw_spinlock_t raw_lock;
} spinlock_t;

struct list_head {
    struct list_head *next;
    struct list_head *prev;
};

struct rcu_head {
    struct rcu_head *next;
    void (*func)(struct rcu_head *);
};

typedef struct {
    int counter;
} atomic_t;

typedef struct rt_prio_array {
    long unsigned int bitmap[2];
    struct list_head queue[100];
} prio_array_t;

typedef struct {
    long unsigned int bits[1];
} cpumask_t;

typedef int pid_t;

struct hlist_node {
    struct hlist_node *next;
    struct hlist_node **pprev;
};

struct hlist_head {
    struct hlist_node *first;
};

struct kref {
    atomic_t refcount;
};

struct upid {
    int nr;
    struct pid_namespace *ns;
    struct hlist_node pid_chain;
};

struct pid {
    atomic_t count;
    unsigned int level;
    struct hlist_head tasks[3];
    struct rcu_head rcu;
    struct upid numbers[1];
};

typedef long unsigned int cputime_t;

typedef long int time_t;

struct kern_timespec {
  time_t tv_sec;
  long int tv_nsec;
};

typedef long loff_t;
typedef unsigned int uid_t;
typedef unsigned int gid_t;
typedef unsigned int __u32;

typedef struct kernel_cap_struct {
    __u32 cap[2];
} kernel_cap_t;

struct sem_undo_list {
    atomic_t refcnt;
    spinlock_t lock;
    struct list_head list_proc;
};

struct sysv_sem {
    struct sem_undo_list *undo_list;
};

typedef short unsigned int u16;
typedef unsigned int u32;
typedef long long unsigned int u64;

typedef struct {
  long unsigned int seg;
} mm_segment_t;

struct restart_block {
    long int (*fn)(struct restart_block *);
    union {
        struct {};
        struct {} futex;
        struct {} nanosleep;
    };
};

struct thread_info {
    struct task_struct *task;
    struct exec_domain *exec_domain;
    long unsigned int flags;
    __u32 status;
    __u32 cpu;
    int preempt_count;
    mm_segment_t addr_limit;
    struct restart_block restart_block;
    void *sysenter_return;
};

struct i387_fxsave_struct {
    u16 cwd;
    u16 swd;
    u16 twd;
    u16 fop;
    union {
        struct {};
        struct {};
    };
    u32 mxcsr;
    u32 mxcsr_mask;
    u32 st_space[32];
    u32 xmm_space[64];
    u32 padding[24];
};

/*
union i387_union {
  struct i387_fxsave_struct fxsave;
};
*/

struct desc_struct {
    union {
        struct task_struct *next;
        struct task_struct *prev;
    };
};

struct thread_struct {
    struct desc_struct tls_array[3];
    long unsigned int sp0;
    long unsigned int sp;
    long unsigned int usersp;
    short unsigned int es;
    short unsigned int ds;
    short unsigned int fsindex;
    short unsigned int gsindex;
    long unsigned int ip;
    long unsigned int fs;
    long unsigned int gs;
    long unsigned int debugreg0;
    long unsigned int debugreg1;
    long unsigned int debugreg2;
    long unsigned int debugreg3;
    long unsigned int debugreg6;
    long unsigned int debugreg7;
    long unsigned int cr2;
    long unsigned int trap_no;
    long unsigned int error_code;
    union thread_xstate *xstate;
    long unsigned int *io_bitmap_ptr;
    long unsigned int iopl;
    unsigned int io_bitmap_max;
    long unsigned int debugctlmsr;
    long unsigned int ds_area_msr;
};

#if 1
typedef struct {
  long unsigned int sig[1];
} kern_sigset_t;
#endif


struct sigpending {
    struct list_head list;
    kern_sigset_t signal;
};

typedef struct {
    int mode;
} seccomp_t;

typedef long int clock_t;

typedef long unsigned int size_t;

typedef struct __wait_queue wait_queue_t;
typedef int (*wait_queue_func_t)(wait_queue_t *, unsigned int, int, void *);

struct __wait_queue {
    unsigned int flags;
    void *private;
    wait_queue_func_t func;
    struct list_head task_list;
} ;

struct load_weight {
    long unsigned int weight;
    long unsigned int inv_weight;
};

struct rb_node {
    long unsigned int rb_parent_color;
    struct rb_node *rb_right;
    struct rb_node *rb_left;
};

typedef long long int s64;
typedef signed char s8;

struct sched_entity {
    struct load_weight load;
    struct rb_node run_node;
    struct list_head group_node;
    unsigned int on_rq;
    u64 exec_start;
    u64 sum_exec_runtime;
    u64 vruntime;
    u64 prev_sum_exec_runtime;
    u64 last_wakeup;
    u64 avg_overlap;
    u64 wait_start;
    u64 wait_max;
    u64 wait_count;
    u64 wait_sum;
    u64 sleep_start;
    u64 sleep_max;
    s64 sum_sleep_runtime;
    u64 block_start;
    u64 block_max;
    u64 exec_max;
    u64 slice_max;
    u64 nr_migrations;
    u64 nr_migrations_cold;
    u64 nr_failed_migrations_affine;
    u64 nr_failed_migrations_running;
    u64 nr_failed_migrations_hot;
    u64 nr_forced_migrations;
    u64 nr_forced2_migrations;
    u64 nr_wakeups;
    u64 nr_wakeups_sync;
    u64 nr_wakeups_migrate;
    u64 nr_wakeups_local;
    u64 nr_wakeups_remote;
    u64 nr_wakeups_affine;
    u64 nr_wakeups_affine_attempts;
    u64 nr_wakeups_passive;
    u64 nr_wakeups_idle;
    struct sched_entity *parent;
    struct cfs_rq *cfs_rq;
    struct cfs_rq *my_q;
};

struct sched_rt_entity {
    struct list_head run_list;
    unsigned int time_slice;
    long unsigned int timeout;
    int nr_cpus_allowed;
    struct sched_rt_entity *back;
    struct sched_rt_entity *parent;
    struct rt_rq *rt_rq;
    struct rt_rq *my_q;
};

struct plist_head {
    struct list_head prio_list;
    struct list_head node_list;
};

struct sched_info {
    long unsigned int pcount;
    long long unsigned int cpu_time;
    long long unsigned int run_delay;
    long long unsigned int last_arrival;
    long long unsigned int last_queued;
    unsigned int bkl_count;
};

struct pid_link {
    struct hlist_node node;
    struct pid *pid;
};

struct task_io_accounting {
    u64 rchar;
    u64 wchar;
    u64 syscr;
    u64 syscw;
    u64 read_bytes;
    u64 write_bytes;
    u64 cancelled_write_bytes;
};

typedef struct {
    long unsigned int bits[8];
} nodemask_t;

struct prop_local_single {
    long unsigned int events;
    int shift;
    long unsigned int period;
    spinlock_t lock;
};

typedef struct {
    unsigned int lock;
} raw_rwlock_t;

typedef struct {
    raw_rwlock_t raw_lock;
} rwlock_t;

struct path {
    struct vfsmount *mnt;
    struct dentry *dentry;
};

struct fs_struct {
    atomic_t count;
    rwlock_t lock;
    int umask;
    struct path root;
    struct path pwd;
};

typedef struct kern_siginfo {
    int si_signo;
    int si_errno;
    int si_code;
    union {
        int _pad[28];
        struct {} _kill;
        struct {} _timer;
        struct {} _rt;
        struct {} _sigchld;
        struct {} _sigfault;
        struct {} _sigpoll;
    } _sifields;
} kern_siginfo_t;

struct latency_record {
    long unsigned int backtrace[12];
    unsigned int count;
    long unsigned int time;
    long unsigned int max;
};

struct rb_root {
    struct rb_node *rb_node;
};

typedef struct {
    long int counter;
} atomic_long_t;

enum pid_type {PIDTYPE_PID, PIDTYPE_PGID, PIDTYPE_SID, PIDTYPE_MAX};

struct fown_struct {
    rwlock_t lock;
    struct pid *pid;
    enum pid_type pid_type;
    uid_t uid;
    uid_t euid;
    int signum;
};

struct file_ra_state {
    long unsigned int start;
    unsigned int size;
    unsigned int async_size;
    unsigned int ra_pages;
    int mmap_miss;
    loff_t prev_pos;
};

typedef long unsigned int pgdval_t;

typedef struct {
    pgdval_t pgd;
} pgd_t;

typedef unsigned int mode_t;

struct file {
    union {
        struct list_head fu_list;
        struct rcu_head fu_rcuhead;
    } f_u;
    struct path f_path;
    const struct file_operations *f_op;
    atomic_long_t f_count;
    unsigned int f_flags;
    mode_t f_mode;
    loff_t f_pos;
    struct fown_struct f_owner;
    unsigned int f_uid;
    unsigned int f_gid;
    struct file_ra_state f_ra;
    u64 f_version;
    void *f_security;
    void *private_data;
    struct list_head f_ep_links;
    spinlock_t f_ep_lock;
    struct address_space *f_mapping;
};

typedef int __s32;

struct rw_semaphore {
    __s32 activity;
    spinlock_t wait_lock;
    struct list_head wait_list;
};

typedef struct {
    long int counter;
} mm_counter_t;

struct mutex {
    atomic_t count;
    spinlock_t wait_lock;
    struct list_head wait_list;
};

struct seq_file {
  char *buf;
  size_t size;
  size_t from;
  size_t count;
  loff_t index;
  loff_t read_ops;
  u64 version;
  struct mutex lock;
  const struct seq_operations *op;
  void *private;
};

typedef struct {
    void *ldt;
    rwlock_t ldtlock;
    int size;
    struct mutex lock;
    void *vdso;
} mm_context_t;

typedef long unsigned int pgprotval_t;

typedef struct {
  pgprotval_t pgprot;
}pgprot_t;

struct prio_tree_node{
  struct prio_tree_node *left;
  struct prio_tree_node *right;
  struct prio_tree_node *parent;
  long unsigned int start;
  long unsigned int last;
};
  

struct raw_prio_tree_node{
  struct prio_tree_node *left;
  struct prio_tree_node *right;
  struct prio_tree_node *parent;
};

struct vm_area_struct{
  struct mm_struct *vm_mm;
  long unsigned int vm_start;
  long unsigned int vm_end;
  struct vm_area_struct *vm_next;
  pgprot_t vm_page_prot;
  long unsigned int vm_flags;
  struct rb_node vm_rb;
  union{
    struct {} vm_set;
    struct raw_prio_tree_node prio_tree_node;
  }shared;
  struct list_head anon_vma_node;
  struct vm_operation_struct *vm_ops;
  long unsigned int vm_pgoff;
  struct file *vm_file;
  void *vm_private_data;
  long unsigned int vm_truncate_count;
  struct mempolicy *vm_policy;
};


struct mm_struct {
    struct vm_area_struct *mmap;
    struct rb_root mm_rb;
    struct vm_area_struct *mmap_cache;
    long unsigned int (*get_unmapped_area)(struct file *, long unsigned int, 
    long unsigned int, long unsigned int, long unsigned int);
    void (*unmap_area)(struct mm_struct *, long unsigned int);
    long unsigned int mmap_base;
    long unsigned int task_size;
    long unsigned int cached_hole_size;
    long unsigned int free_area_cache;
    pgd_t *pgd;
    atomic_t mm_users;
    atomic_t mm_count;
    int map_count;
    struct rw_semaphore mmap_sem;
    spinlock_t page_table_lock;
    struct list_head mmlist;
    mm_counter_t _file_rss;
    mm_counter_t _anon_rss;
    long unsigned int hiwater_rss;
    long unsigned int hiwater_vm;
    long unsigned int total_vm;
    long unsigned int locked_vm;
    long unsigned int shared_vm;
    long unsigned int exec_vm;
    long unsigned int stack_vm;
    long unsigned int reserved_vm;
    long unsigned int def_flags;
    long unsigned int nr_ptes;
    long unsigned int start_code;
    long unsigned int end_code;
    long unsigned int start_data;
    long unsigned int end_data;
    long unsigned int start_brk;
    long unsigned int brk;
    long unsigned int start_stack;
    long unsigned int arg_start;
    long unsigned int arg_end;
    long unsigned int env_start;
    long unsigned int env_end;
    long unsigned int saved_auxv[42];
    cpumask_t cpu_vm_mask;
    mm_context_t context;
    unsigned int faultstamp;
    unsigned int token_priority;
    unsigned int last_interval;
    long unsigned int flags;
    struct core_state *core_state;
    rwlock_t ioctx_list_lock;
    struct kioctx *ioctx_list;
    struct file *exe_file;
    long unsigned int num_exe_file_vmas;
    struct mmu_notifier_mm *mmu_notifier_mm;
};

typedef void (*__sighandler_t)(int);
typedef void (*__sigrestore_t)(void);

struct kern_sigaction {
    __sighandler_t sa_handler;
    long unsigned int sa_flags;
    __sigrestore_t sa_restorer;
    kern_sigset_t sa_mask;
};

struct k_sigaction {
    struct kern_sigaction sa;
};

typedef struct __wait_queue_head {
    spinlock_t lock;
    struct list_head task_list;
} wait_queue_head_t;

struct sighand_struct {
    atomic_t count;
    struct k_sigaction action[64];
    spinlock_t siglock;
    wait_queue_head_t signalfd_wqh;
};
struct user_struct {
    atomic_t __count;
    atomic_t processes;
    atomic_t files;
    atomic_t sigpending;
    atomic_t inotify_watches;
    atomic_t inotify_devs;
    atomic_t epoll_watches;
    long unsigned int mq_bytes;
    long unsigned int locked_shm;
    struct key *uid_keyring;
    struct key *session_keyring;
    struct hlist_node uidhash_node;
    uid_t uid;
};

struct tty_ldisc {
    struct tty_ldisc_ops *ops;
    int refcount;
};

struct winsize {
    short unsigned int ws_row;
    short unsigned int ws_col;
    short unsigned int ws_xpixel;
    short unsigned int ws_ypixel;
};

struct timer_list {
    struct list_head entry;
    long unsigned int expires;
    void (*function)(long unsigned int);
    long unsigned int data;
    struct tvec_base *base;
    void *start_site;
    char start_comm[16];
    int start_pid;
};

struct work_struct;

typedef void (*work_func_t)(struct work_struct *);

struct work_struct {
    atomic_long_t data;
    struct list_head entry;
    work_func_t func;
};

struct delayed_work {
    struct work_struct work;
    struct timer_list timer;
};

struct tty_bufhead {
    struct delayed_work work;
    spinlock_t lock;
    struct tty_buffer *head;
    struct tty_buffer *tail;
    struct tty_buffer *free;
    int memory_used;
};

struct kobject {
    const char *name;
    struct list_head entry;
    struct kobject *parent;
    struct kset *kset;
    struct kobj_type *ktype;
    struct sysfs_dirent *sd;
    struct kref kref;
    unsigned int state_initialized : 1;
    unsigned int state_in_sysfs : 1;
    unsigned int state_add_uevent_sent : 1;
    unsigned int state_remove_uevent_sent : 1;
};

typedef unsigned int kern_dev_t;

struct cdev {
    struct kobject kobj;
    struct module *owner;
    const struct file_operations *ops;
    struct list_head list;
    kern_dev_t dev;
    unsigned int count;
};

typedef unsigned int tcflag_t;
typedef unsigned char cc_t;
typedef unsigned int speed_t;

struct ktermios {
    tcflag_t c_iflag;
    tcflag_t c_oflag;
    tcflag_t c_cflag;
    tcflag_t c_lflag;
    cc_t c_line;
    cc_t c_cc[19];
    speed_t c_ispeed;
    speed_t c_ospeed;
};

enum module_state {MODULE_STATE_LIVE, MODULE_STATE_COMING, MODULE_STATE_GOING};

struct module_kobject {
  struct kobject kobj;
  struct module *mod;
  struct kobject *drivers_dir;
};

struct mod_arch_specific {};
typedef unsigned int Elf64_Word;
typedef short unsigned int Elf64_Half;
typedef long long unsigned int Elf64_Addr;
typedef long long unsigned int Elf64_Xword;

typedef struct elf64_sym {
  Elf64_Word st_name;
  unsigned char st_info;
  unsigned char st_other;
  Elf64_Half st_shndx;
  Elf64_Addr st_value;
  Elf64_Xword st_size;
} Elf64_Sym;

typedef struct {
  atomic_long_t a;
} local_t;

struct module_ref {
  local_t count;
};

struct module {
    enum module_state state;
    struct list_head list;
    char name[56];
    struct module_kobject mkobj;
    struct module_param_attrs *param_attrs;
    struct module_attribute *modinfo_attrs;
    const char *version;
    const char *srcversion;
    struct kobject *holders_dir;
    const struct kernel_symbol *syms;
    const long unsigned int *crcs;
    unsigned int num_syms;
    unsigned int num_gpl_syms;
    const struct kernel_symbol *gpl_syms;
    const long unsigned int *gpl_crcs;
    const struct kernel_symbol *unused_syms;
    const long unsigned int *unused_crcs;
    unsigned int num_unused_syms;
    unsigned int num_unused_gpl_syms;
    const struct kernel_symbol *unused_gpl_syms;
    const long unsigned int *unused_gpl_crcs;
    const struct kernel_symbol *gpl_future_syms;
    const long unsigned int *gpl_future_crcs;
    unsigned int num_gpl_future_syms;
    unsigned int num_exentries;
    const struct exception_table_entry *extable;
    int (*init)(void);
    void *module_init;
    void *module_core;
    unsigned int init_size;
    unsigned int core_size;
    unsigned int init_text_size;
    unsigned int core_text_size;
    void *unwind_info;
    struct mod_arch_specific arch;
    unsigned int taints;
    unsigned int num_bugs;
    struct list_head bug_list;
    struct bug_entry *bug_table;
    Elf64_Sym *symtab;
    unsigned int num_symtab;
    char *strtab;
    struct module_sect_attrs *sect_attrs;
    struct module_notes_attrs *notes_attrs;
    void *percpu;
    char *args;
    struct marker *markers;
    unsigned int num_markers;
    struct list_head modules_which_use_me;
    struct task_struct *waiter;
    void (*exit)(void);
    struct module_ref ref[64];
};

struct tty_driver {
    int magic;
    struct cdev cdev;
    struct module *owner;
    const char *driver_name;
    const char *name;
    int name_base;
    int major;
    int minor_start;
    int minor_num;
    int num;
    short int type;
    short int subtype;
    struct ktermios init_termios;
    int flags;
    int refcount;
    struct proc_dir_entry *proc_entry;
    struct tty_driver *other;
    struct tty_struct **ttys;
    struct ktermios **termios;
    struct ktermios **termios_locked;
    void *driver_state;
    const struct tty_operations *ops;
    struct list_head tty_drivers;
};

struct tty_struct {
    int magic;
    struct tty_driver *driver;
    const struct tty_operations *ops;
    int index;
    struct tty_ldisc ldisc;
    struct mutex termios_mutex;
    spinlock_t ctrl_lock;
    struct ktermios *termios;
    struct ktermios *termios_locked;
    char name[64];
    struct pid *pgrp;
    struct pid *session;
    long unsigned int flags;
    int count;
    struct winsize winsize;
    unsigned char stopped : 1;
    unsigned char hw_stopped : 1;
    unsigned char flow_stopped : 1;
    unsigned char packet : 1;
    unsigned char low_latency : 1;
    unsigned char warned : 1;
    unsigned char ctrl_status;
    unsigned int receive_room;
    struct tty_struct *link;
    struct fasync_struct *fasync;
    struct tty_bufhead buf;
    int alt_speed;
    wait_queue_head_t write_wait;
    wait_queue_head_t read_wait;
    struct work_struct hangup_work;
    void *disc_data;
    void *driver_data;
    struct list_head tty_files;
    unsigned int column;
    unsigned char lnext : 1;
    unsigned char erasing : 1;
    unsigned char raw : 1;
    unsigned char real_raw : 1;
    unsigned char icanon : 1;
    unsigned char closing : 1;
    short unsigned int minimum_to_wake;
    long unsigned int overrun_time;
    int num_overrun;
    long unsigned int process_char_map[4];
    char *read_buf;
    int read_head;
    int read_tail;
    int read_cnt;
    long unsigned int read_flags[64];
    int canon_data;
    long unsigned int canon_head;
    unsigned int canon_column;
    struct mutex atomic_read_lock;
    struct mutex atomic_write_lock;
    unsigned char *write_buf;
    int write_cnt;
    spinlock_t read_lock;
    struct work_struct SAK_work;
    struct tty_port *port;
};

struct task_delay_info {
    spinlock_t lock;
    unsigned int flags;
    struct kern_timespec blkio_start;
    struct kern_timespec blkio_end;
    u64 blkio_delay;
    u64 swapin_delay;
    u32 blkio_count;
    u32 swapin_count;
    struct kern_timespec freepages_start;
    struct kern_timespec freepages_end;
    u64 freepages_delay;
    u32 freepages_count;
};

struct new_utsname {
    char sysname[65];
    char nodename[65];
    char release[65];
    char version[65];
    char machine[65];
    char domainname[65];
};

struct uts_namespace {
    struct kref kref;
    struct new_utsname name;
};

struct nsproxy {
    atomic_t count;
    struct uts_namespace *uts_ns;
    struct ipc_namespace *ipc_ns;
    struct mnt_namespace *mnt_ns;
    struct pid_namespace *pid_ns;
    struct user_namespace *user_ns;
    struct net *net_ns;
};

typedef struct task_struct {
    volatile long int state;
    void *stack;
    atomic_t usage;
    unsigned int flags;
    unsigned int ptrace;
    int lock_depth;
    int prio;
    int static_prio;
    int normal_prio;
    unsigned int rt_priority;
    const struct sched_class *sched_class;
    struct sched_entity se;
    struct sched_rt_entity rt;
    struct hlist_head preempt_notifiers;
    unsigned char fpu_counter;
    s8 oomkilladj;
    unsigned int btrace_seq;
    unsigned int policy;
    cpumask_t cpus_allowed;
    struct sched_info sched_info;
    struct list_head tasks;
    struct mm_struct *mm;
    struct mm_struct *active_mm;
    struct linux_binfmt *binfmt;
    int exit_state;
    int exit_code;
    int exit_signal;
    int pdeath_signal;
    unsigned int personality;
    unsigned int did_exec : 1;
    pid_t pid;
    pid_t tgid;
    struct task_struct *real_parent;
    struct task_struct *parent;
    struct list_head children;
    struct list_head sibling;
    struct task_struct *group_leader;
    struct list_head ptraced;
    struct list_head ptrace_entry;
    struct pid_link pids[3];
    struct list_head thread_group;
    struct completion *vfork_done;
    int *set_child_tid;
    int *clear_child_tid;
    cputime_t utime;
    cputime_t stime;
    cputime_t utimescaled;
    cputime_t stimescaled;
    cputime_t gtime;
    cputime_t prev_utime;
    cputime_t prev_stime;
    long unsigned int nvcsw;
    long unsigned int nivcsw;
    struct kern_timespec start_time;
    struct kern_timespec real_start_time;
    long unsigned int min_flt;
    long unsigned int maj_flt;
    cputime_t it_prof_expires;
    cputime_t it_virt_expires;
    long long unsigned int it_sched_expires;
    struct list_head cpu_timers[3];
    uid_t uid;
    uid_t euid;
    uid_t suid;
    uid_t fsuid;
    gid_t gid;
    gid_t egid;
    gid_t sgid;
    gid_t fsgid;
    struct group_info *group_info;
    kernel_cap_t cap_effective;
    kernel_cap_t cap_inheritable;
    kernel_cap_t cap_permitted;
    kernel_cap_t cap_bset;
    struct user_struct *user;
    unsigned int securebits;
    unsigned char jit_keyring;
    struct key *request_key_auth;
    struct key *thread_keyring;
    char comm[16];
    int link_count;
    int total_link_count;
    struct sysv_sem sysvsem;
    long unsigned int last_switch_timestamp;
    long unsigned int last_switch_count;
    struct thread_struct thread;
    struct fs_struct *fs;
    struct files_struct *files;
    struct nsproxy *nsproxy;
    struct signal_struct *signal;
    struct sighand_struct *sighand;
    kern_sigset_t blocked;
    kern_sigset_t real_blocked;
    kern_sigset_t saved_sigmask;
    struct sigpending pending;
    long unsigned int sas_ss_sp;
    size_t sas_ss_size;
    int (*notifier)(void *);
    void *notifier_data;
    kern_sigset_t *notifier_mask;
    void *security;
    struct audit_context *audit_context;
    uid_t loginuid;
    unsigned int sessionid;
    seccomp_t seccomp;
    u32 parent_exec_id;
    u32 self_exec_id;
    spinlock_t alloc_lock;
    spinlock_t pi_lock;
    struct plist_head pi_waiters;
    struct rt_mutex_waiter *pi_blocked_on;
    unsigned int irq_events;
    int hardirqs_enabled;
    long unsigned int hardirq_enable_ip;
    unsigned int hardirq_enable_event;
    long unsigned int hardirq_disable_ip;
    unsigned int hardirq_disable_event;
    int softirqs_enabled;
    long unsigned int softirq_disable_ip;
    unsigned int softirq_disable_event;
    long unsigned int softirq_enable_ip;
    unsigned int softirq_enable_event;
    int hardirq_context;
    int softirq_context;
    void *journal_info;
    struct bio *bio_list;
    struct bio **bio_tail;
    struct reclaim_state *reclaim_state;
    struct backing_dev_info *backing_dev_info;
    struct io_context *io_context;
    long unsigned int ptrace_message;
    void *last_siginfo; //changed siginto_t
    struct task_io_accounting ioac;
    u64 acct_rss_mem1;
    u64 acct_vm_mem1;
    cputime_t acct_timexpd;
    nodemask_t mems_allowed;
    int cpuset_mems_generation;
    int cpuset_mem_spread_rotor;
    struct css_set *cgroups;
    struct list_head cg_list;
    struct robust_list_head *robust_list;
    struct compat_robust_list_head *compat_robust_list;
    struct list_head pi_state_list;
    struct futex_pi_state *pi_state_cache;
    struct mempolicy *mempolicy;
    short int il_next;
    atomic_t fs_excl;
    struct rcu_head rcu;
    struct list_head *scm_work_list;
    struct pipe_inode_info *splice_pipe;
    struct task_delay_info *delays;
    struct prop_local_single dirties;
    int latency_record_count;
    struct latency_record latency_record[32];
} task_t;

struct embedded_fd_set {
    long unsigned int fds_bits[1];
};

typedef struct {
    long unsigned int fds_bits[16];
} kern_fd_set;

struct fdtable {
    unsigned int max_fds;
    struct file **fd;
    kern_fd_set *close_on_exec;
    kern_fd_set *open_fds;
    struct rcu_head rcu;
    struct fdtable *next;
};

struct files_struct {
    atomic_t count;
    struct fdtable *fdt;
    struct fdtable fdtab;
    spinlock_t file_lock;
    int next_fd;
    struct embedded_fd_set close_on_exec_init;
    struct embedded_fd_set open_fds_init;
    struct file *fd_array[64];
};

struct cfs_rq {
    struct load_weight load;
    long unsigned int nr_running;
    u64 exec_clock;
    u64 min_vruntime;
    u64 pair_start;
    struct rb_root tasks_timeline;
    struct rb_node *rb_leftmost;
    struct list_head tasks;
    struct list_head *balance_iterator;
    struct sched_entity *curr;
    struct sched_entity *next;
    long unsigned int nr_spread_over;
};

struct rt_rq {
    struct rt_prio_array active;
    long unsigned int rt_nr_running;
    int highest_prio;
    long unsigned int rt_nr_migratory;
    int overloaded;
    int rt_throttled;
    u64 rt_time;
    u64 rt_runtime;
    spinlock_t rt_runtime_lock;
    long unsigned int rt_nr_boosted;
    struct rq *rq;
    struct list_head leaf_rt_rq_list;
    struct task_group *tg;
    struct sched_rt_entity *rt_se;
};

struct group_info {
    int ngroups;
    atomic_t usage;
    gid_t small_block[32];
    int nblocks;
    gid_t *blocks[];
};

struct call_single_data {
    struct list_head list;
    void (*func)(void *);
    void *info;
    unsigned int flags;
};

typedef union ktime {
    s64 tv64;
} ktime_t;

enum hrtimer_cb_mode {HRTIMER_CB_SOFTIRQ, HRTIMER_CB_IRQSAFE, 
    HRTIMER_CB_IRQSAFE_NO_RESTART, HRTIMER_CB_IRQSAFE_PERCPU, 
		      HRTIMER_CB_IRQSAFE_UNLOCKED};

struct hrtimer {
    struct rb_node node;
    ktime_t expires;
    enum hrtimer_restart (*function)(struct hrtimer *);
    struct hrtimer_clock_base *base;
    long unsigned int state;
    enum hrtimer_cb_mode cb_mode;
    struct list_head cb_entry;
    void *start_site;
    char start_comm[16];
    int start_pid;
};


typedef struct rq {
    spinlock_t lock;
    long unsigned int nr_running;
    long unsigned int cpu_load[5];
    unsigned char idle_at_tick;
    long unsigned int last_tick_seen;
    unsigned char in_nohz_recently;
    struct load_weight load;
    long unsigned int nr_load_updates;
    u64 nr_switches;
    struct cfs_rq cfs;
    struct rt_rq rt;
    struct list_head leaf_cfs_rq_list;
    struct list_head leaf_rt_rq_list;
    long unsigned int nr_uninterruptible;
    struct task_struct *curr;
    struct task_struct *idle;
    long unsigned int next_balance;
    struct mm_struct *prev_mm;
    u64 clock;
    atomic_t nr_iowait;
    struct root_domain *rd;
    struct sched_domain *sd;
    int active_balance;
    int push_cpu;
    int cpu;
    int online;
    long unsigned int avg_load_per_task;
    struct task_struct *migration_thread;
    struct list_head migration_queue;
    int hrtick_csd_pending;
    struct call_single_data hrtick_csd;
    struct hrtimer hrtick_timer;
    struct sched_info rq_sched_info;
    unsigned int yld_exp_empty;
    unsigned int yld_act_empty;
    unsigned int yld_both_empty;
    unsigned int yld_count;
    unsigned int sched_switch;
    unsigned int sched_count;
    unsigned int sched_goidle;
    unsigned int ttwu_count;
    unsigned int ttwu_local;
    unsigned int bkl_count;
} runqueue_t;

struct pidmap {
    atomic_t nr_free;
    void *page;
};

struct pid_namespace {
    struct kref kref;
    struct pidmap pidmap[128];
    int last_pid;
    struct task_struct *child_reaper;
    struct kmem_cache *pid_cachep;
    unsigned int level;
    struct pid_namespace *parent;
    struct vfsmount *proc_mnt;
    struct bsd_acct_struct *bacct;
};

#undef offsetof
#ifdef __compiler_offsetof
#define offsetof(TYPE,MEMBER) __compiler_offsetof(TYPE,MEMBER)
#else
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif
#define __compiler_offsetof(a,b) __builtin_offsetof(a,b)

#define MAX_USER_RT_PRIO        100
#define MAX_RT_PRIO             MAX_USER_RT_PRIO
#define MAX_PRIO                (MAX_RT_PRIO + 40)

struct x8664_pda {
    struct task_struct *pcurrent;
    long unsigned int data_offset;
    long unsigned int kernelstack;
    long unsigned int oldrsp;
    int irqcount;
    unsigned int cpunumber;
    char *irqstackptr;
    short int nodenumber;
    short int in_bootmem;
    unsigned int __softirq_pending;
    unsigned int __nmi_count;
    short int mmu_state;
    short int isidle;
    struct mm_struct *active_mm;
    unsigned int apic_timer_irqs;
    unsigned int irq0_irqs;
    unsigned int irq_resched_count;
    unsigned int irq_call_count;
    unsigned int irq_tlb_count;
    unsigned int irq_thermal_count;
    unsigned int irq_threshold_count;
    unsigned int irq_spurious_count;
};

struct semaphore {
    spinlock_t lock;
    unsigned int count;
    struct list_head wait_list;
};

struct completion {
    unsigned int done;
    wait_queue_head_t wait;
};

struct kern_rlimit {
    long unsigned int rlim_cur;
    long unsigned int rlim_max;
};

struct pacct_struct {
    int ac_flag;
    long int ac_exitcode;
    long unsigned int ac_mem;
    cputime_t ac_utime;
    cputime_t ac_stime;
    long unsigned int ac_minflt;
    long unsigned int ac_majflt;
};

struct signal_struct {
    atomic_t count;
    atomic_t live;
    wait_queue_head_t wait_chldexit;
    struct task_struct *curr_target;
    struct sigpending shared_pending;
    int group_exit_code;
    struct task_struct *group_exit_task;
    int notify_count;
    int group_stop_count;
    unsigned int flags;
    struct list_head posix_timers;
    struct hrtimer real_timer;
    struct pid *leader_pid;
    ktime_t it_real_incr;
    cputime_t it_prof_expires;
    cputime_t it_virt_expires;
    cputime_t it_prof_incr;
    cputime_t it_virt_incr;
    union {
        pid_t pgrp;
        pid_t __pgrp;
    };
    struct pid *tty_old_pgrp;
    union {
        pid_t session;
        pid_t __session;
    };
    int leader;
    struct tty_struct *tty;
    cputime_t utime;
    cputime_t stime;
    cputime_t cutime;
    cputime_t cstime;
    cputime_t gtime;
    cputime_t cgtime;
    long unsigned int nvcsw;
    long unsigned int nivcsw;
    long unsigned int cnvcsw;
    long unsigned int cnivcsw;
    long unsigned int min_flt;
    long unsigned int maj_flt;
    long unsigned int cmin_flt;
    long unsigned int cmaj_flt;
    long unsigned int inblock;
    long unsigned int oublock;
    long unsigned int cinblock;
    long unsigned int coublock;
    struct task_io_accounting ioac;
    long long unsigned int sum_sched_runtime;
    struct kern_rlimit rlim[16];
    struct list_head cpu_timers[3];
    struct key *session_keyring;
    struct key *process_keyring;
    struct pacct_struct pacct;
    struct taskstats *stats;
    unsigned int audit_tty;
    struct tty_audit_buf *tty_audit_buf;
};

typedef long unsigned int pteval_t;

typedef struct {
    pteval_t pte;
} pte_t;

struct klist_node {
    struct klist *n_klist;
    struct list_head n_node;
    struct kref n_ref;
    struct completion n_removed;
};

struct klist {
    spinlock_t k_lock;
    struct list_head k_list;
    void (*get)(struct klist_node *);
    void (*put)(struct klist_node *);
};

typedef struct pm_message {
  int event;
} pm_message_t;

enum dpm_state {DPM_INVALID, DPM_ON, DPM_PREPARING, DPM_RESUMING, 
		DPM_SUSPENDING, DPM_OFF, DPM_OFF_IRQ};

struct dev_pm_info {
    pm_message_t power_state;
    unsigned int can_wakeup : 1;
    unsigned int should_wakeup : 1;
    enum dpm_state status;
    struct list_head entry;
};

struct dev_archdata {
  void *acpi_handle;
  struct dma_mapping_ops *dma_ops;
};

struct device {
    struct klist klist_children;
    struct klist_node knode_parent;
    struct klist_node knode_driver;
    struct klist_node knode_bus;
    struct device *parent;
    struct kobject kobj;
    char bus_id[20];
    const char *init_name;
    struct device_type *type;
    unsigned int uevent_suppress : 1;
    struct semaphore sem;
    struct bus_type *bus;
    struct device_driver *driver;
    void *driver_data;
    void *platform_data;
    struct dev_pm_info power;
    u64 *dma_mask;
    u64 coherent_dma_mask;
    struct device_dma_parameters *dma_parms;
    struct list_head dma_pools;
    struct dma_coherent_mem *dma_mem;
    struct dev_archdata archdata;
    spinlock_t devres_lock;
    struct list_head devres_head;
    struct list_head node;
    struct class *class;
    kern_dev_t devt;
    struct attribute_group **groups;
    void (*release)(struct device *);
};

struct sock_common {
    short unsigned int skc_family;
    volatile unsigned char skc_state;
    unsigned char skc_reuse;
    int skc_bound_dev_if;
    struct hlist_node skc_node;
    struct hlist_node skc_bind_node;
    atomic_t skc_refcnt;
    unsigned int skc_hash;
    struct proto *skc_prot;
};

typedef struct {
    spinlock_t slock;
    int owned;
    wait_queue_head_t wq;
} socket_lock_t;

struct sk_buff_head {
    struct sk_buff *next;
    struct sk_buff *prev;
    __u32 qlen;
    spinlock_t lock;
};

typedef unsigned int gfp_t;

struct ucred {
    __u32 pid;
    __u32 uid;
    __u32 gid;
};

struct sock {
    struct sock_common __sk_common;
#define sk_family               __sk_common.skc_family
#define sk_state                __sk_common.skc_state
#define sk_reuse                __sk_common.skc_reuse
#define sk_bound_dev_if         __sk_common.skc_bound_dev_if
#define sk_node                 __sk_common.skc_node
#define sk_bind_node            __sk_common.skc_bind_node
#define sk_refcnt               __sk_common.skc_refcnt
#define sk_hash                 __sk_common.skc_hash
#define sk_prot                 __sk_common.skc_prot
#define sk_net                  __sk_common.skc_net
    unsigned char sk_shutdown : 2;
    unsigned char sk_no_check : 2;
    unsigned char sk_userlocks : 4;
    unsigned char sk_protocol;
    short unsigned int sk_type;
    int sk_rcvbuf;
    socket_lock_t sk_lock;
    struct {
        struct sk_buff *head;
        struct sk_buff *tail;
    } sk_backlog;
    wait_queue_head_t *sk_sleep;
    struct dst_entry *sk_dst_cache;
    struct xfrm_policy *sk_policy[2];
    rwlock_t sk_dst_lock;
    atomic_t sk_rmem_alloc;
    atomic_t sk_wmem_alloc;
    atomic_t sk_omem_alloc;
    int sk_sndbuf;
    struct sk_buff_head sk_receive_queue;
    struct sk_buff_head sk_write_queue;
    struct sk_buff_head sk_async_wait_queue;
    int sk_wmem_queued;
    int sk_forward_alloc;
    gfp_t sk_allocation;
    int sk_route_caps;
    int sk_gso_type;
    unsigned int sk_gso_max_size;
    int sk_rcvlowat;
    long unsigned int sk_flags;
    long unsigned int sk_lingertime;
    struct sk_buff_head sk_error_queue;
    struct proto *sk_prot_creator;
    rwlock_t sk_callback_lock;
    int sk_err;
    int sk_err_soft;
    atomic_t sk_drops;
    short unsigned int sk_ack_backlog;
    short unsigned int sk_max_ack_backlog;
    __u32 sk_priority;
    struct ucred sk_peercred;
    long int sk_rcvtimeo;
    long int sk_sndtimeo;
    struct sk_filter *sk_filter;
    void *sk_protinfo;
    struct timer_list sk_timer;
    ktime_t sk_stamp;
    struct socket *sk_socket;
    void *sk_user_data;
    struct page *sk_sndmsg_page;
    struct sk_buff *sk_send_head;
    __u32 sk_sndmsg_off;
    int sk_write_pending;
    void *sk_security;
    __u32 sk_mark;
    void (*sk_state_change)(struct sock *);
    void (*sk_data_ready)(struct sock *, int);
    void (*sk_write_space)(struct sock *);
    void (*sk_error_report)(struct sock *);
    int (*sk_backlog_rcv)(struct sock *, struct sk_buff *);
    void (*sk_destruct)(struct sock *);
};

typedef short unsigned int sa_family_t;

struct sockaddr_un {
    sa_family_t sun_family;
    char sun_path[108];
};

struct unix_address {
    atomic_t refcnt;
    int len;
    unsigned int hash;
    struct sockaddr_un name[0];
};

struct unix_sock {
    struct sock sk;
    struct unix_address *addr;
    struct dentry *dentry;
    struct vfsmount *mnt;
    struct mutex readlock;
    struct sock *peer;
    struct list_head link;
    atomic_long_t inflight;
    spinlock_t lock;
    unsigned int gc_candidate : 1;
    unsigned int gc_maybe_cycle : 1;
    wait_queue_head_t peer_wait;
};

enum socket_state {SS_FREE, SS_UNCONNECTED, SS_CONNECTING, SS_CONNECTED, SS_DISCONNECTING};

struct socket{
    enum socket_state state;
    short int type;
    long unsigned int flags;
    const struct proto_ops *ops;
    struct fasync_struct *fasync_list;
    struct file *file;
    struct sock *sk;
    wait_queue_head_t wait;
};

typedef short unsigned int umode_t;

struct radix_tree_root {
		unsigned int height;
		gfp_t gfp_mask;
		struct radix_tree_node *rnode;
};

struct prio_tree_root {
		struct prio_tree_node *prio_tree_node;
		short unsigned int index_bits;
		short unsigned int raw;
};

struct address_space {
    struct inode *host;
    struct radix_tree_root page_tree;
    spinlock_t tree_lock;
    unsigned int i_mmap_writable;
    struct prio_tree_root i_mmap;
    struct list_head i_mmap_nonlinear;
    spinlock_t i_mmap_lock;
    unsigned int truncate_count;
    long unsigned int nrpages;
    long unsigned int writeback_index;
    const struct address_space_operations *a_ops;
    long unsigned int flags;
    struct backing_dev_info *backing_dev_info;
    spinlock_t private_lock;
    struct list_head private_list;
    struct address_space *assoc_mapping;
};

typedef long unsigned int kern_blkcnt_t;

struct inode {
    struct hlist_node i_hash;
    struct list_head i_list;
    struct list_head i_sb_list;
    struct list_head i_dentry;
    long unsigned int i_ino;
    atomic_t i_count;
    unsigned int i_nlink;
    uid_t i_uid;
    gid_t i_gid;
    kern_dev_t i_rdev;
    u64 i_version;
    loff_t i_size;
    struct kern_timespec i_atime;
    struct kern_timespec i_mtime;
    struct kern_timespec i_ctime;
    unsigned int i_blkbits;
    kern_blkcnt_t i_blocks;   /*change blkcnt_t*/
    short unsigned int i_bytes;
    umode_t i_mode;
    spinlock_t i_lock;
    struct mutex i_mutex;
    struct rw_semaphore i_alloc_sem;
    const struct inode_operations *i_op;
    const struct file_operations *i_fop;
    struct super_block *i_sb;
    struct file_lock *i_flock;
    struct address_space *i_mapping;
    struct address_space i_data;
    struct dquot *i_dquot[2];
    struct list_head i_devices;
    union {
        struct pipe_inode_info *i_pipe;
        struct block_device *i_bdev;
        struct cdev *i_cdev;
    };
    int i_cindex;
    __u32 i_generation;
    long unsigned int i_dnotify_mask;
    struct dnotify_struct *i_dnotify;
    struct list_head inotify_watches;
    struct mutex inotify_mutex;
    long unsigned int i_state;
    long unsigned int dirtied_when;
    unsigned int i_flags;
    atomic_t i_writecount;
    void *i_security;
    void *i_private;
};

struct socket_alloc {
		struct socket socket;
		struct inode vfs_inode;
};

struct inet_ehash_bucket {
	struct hlist_head chain;
	struct hlist_head twchain;
};

struct inet_hashinfo {
	struct inet_ehash_bucket *ehash;
	rwlock_t *ehash_locks;
	unsigned int ehash_size;
	unsigned int ehahs_locks_mask;
	struct inet_bind_hashbucket *bhash;
	unsigned int bhash_size;
	struct hlist_head listening_hash[32];
	rwlock_t lhash_lock ____cacheline_aligned;
	atomic_t lhash_users;
	wait_queue_head_t lhash_wait;
	struct kmem_cache *bind_bucket_cachep;
};

typedef unsigned int __be32;
typedef short unsigned int __be16;
typedef short unsigned int __u16;
typedef short int __s16;
typedef short unsigned int __le16;
typedef unsigned char __u8;
typedef __u8 u8;

struct in6_addr {
	union {
		__u8 u6_addr8[16];
		__be16 u6_addr16[8];
		__be32 u6_addr32[4];
	} in6_u;
};

struct flowi {
        int     oif;
        int     iif;
        __u32   mark;

        union {
                struct {
                        __be32                  daddr;
                        __be32                  saddr;
                        __u8                    tos;
                        __u8                    scope;
                } ip4_u;
                
                struct {
                        struct in6_addr         daddr;
                        struct in6_addr         saddr;
                        __be32                  flowlabel;
                } ip6_u;

                struct {
                        __le16                  daddr;
                        __le16                  saddr;
                        __u8                    scope;
                } dn_u;
        } nl_u;
#define fld_dst         nl_u.dn_u.daddr
#define fld_src         nl_u.dn_u.saddr
#define fld_scope       nl_u.dn_u.scope
#define fl6_dst         nl_u.ip6_u.daddr
#define fl6_src         nl_u.ip6_u.saddr
#define fl6_flowlabel   nl_u.ip6_u.flowlabel
#define fl4_dst         nl_u.ip4_u.daddr
#define fl4_src         nl_u.ip4_u.saddr
#define fl4_tos         nl_u.ip4_u.tos
#define fl4_scope       nl_u.ip4_u.scope

        __u8    proto;
        union {
                struct {
                        __be16  sport;
                        __be16  dport;
                } ports;

                struct {
                        __u8    type;
                        __u8    code;
                } icmpt;

                struct {
                        __le16  sport;
                        __le16  dport;
                } dnports;

                __be32          spi;

                struct {
                        __u8    type;
                } mht;
        } uli_u;
#define fl_ip_sport     uli_u.ports.sport
#define fl_ip_dport     uli_u.ports.dport
#define fl_icmp_type    uli_u.icmpt.type
#define fl_icmp_code    uli_u.icmpt.code
#define fl_ipsec_spi    uli_u.spi
#define fl_mh_type      uli_u.mht.type
        __u32           secid;  /* used by xfrm; see secid.txt */
} __attribute__((__aligned__(BITS_PER_LONG/8)));

/*
struct flowi {
	int	oif;
	int	iif;
	__u32	mark;

	union {
		struct {
			__be32	daddr;
			__be32	saddr;
			__u8	tos;
			__u8	scope;
		} ip4_u;

		struct {
			struct in6_addr	daddr;
			struct in6_addr	saddr;
			__be32	flowlabel;
		} ip6_u;

		struct {
			__le16	daddr;
			__le16	saddr;
			__u8	scope;
		} dn_u;
	} nl_u;
};
*/

struct inet_sock {
    struct sock sk;
    struct ipv6_pinfo *pinet6;
    __be32 daddr;
    __be32 rcv_saddr;
    __be16 dport;
    __u16 num;
    __be32 saddr;
    __s16 uc_ttl;
    __u16 cmsg_flags;
    struct ip_options *opt;
    __be16 sport;
    __u16 id;
    __u8 tos;
    __u8 mc_ttl;
    __u8 pmtudisc;
    __u8 recverr : 1;
    __u8 is_icsk : 1;
    __u8 freebind : 1;
    __u8 hdrincl : 1;
    __u8 mc_loop : 1;
    int mc_index;
    __be32 mc_addr;
    struct ip_mc_socklist *mc_list;
    struct {
        unsigned int flags;
        unsigned int fragsize;
        struct ip_options *opt;
        struct dst_entry *dst;
        int length;
        __be32 addr;
        struct flowi fl;
    } cork;
};

struct request_sock;

struct request_sock_ops {
    int family;
    int obj_size;
    struct kmem_cache *slab;
    int (*rtx_syn_ack)(struct sock *, struct request_sock *);
    void (*send_ack)(struct sock *, struct sk_buff *, struct request_sock *);
    void (*send_reset)(struct sock *, struct sk_buff *);
    void (*destructor)(struct request_sock *);
};

struct request_sock {
    struct request_sock *dl_next;
    u16 mss;
    u8 retrans;
    u8 cookie_ts;
    u32 window_clamp;
    u32 rcv_wnd;
    u32 ts_recent;
    long unsigned int expires;
    const struct request_sock_ops *rsk_ops;
    struct sock *sk;
    u32 secid;
    u32 peer_secid;
};

struct listen_sock {
    u8 max_qlen_log;
    int qlen;
    int qlen_young;
    int clock_hand;
    u32 hash_rnd;
    u32 nr_table_entries;
    struct request_sock *syn_table[];
};

struct request_sock_queue {
    struct request_sock *rskq_accept_head;
    struct request_sock *rskq_accept_tail;
    rwlock_t syn_wait_lock;
    u8 rskq_defer_accept;
    struct listen_sock *listen_opt;
};

struct inet_connection_sock {
    struct inet_sock icsk_inet;
    struct request_sock_queue icsk_accept_queue;
    struct inet_bind_bucket *icsk_bind_hash;
    long unsigned int icsk_timeout;
    struct timer_list icsk_retransmit_timer;
    struct timer_list icsk_delack_timer;
    __u32 icsk_rto;
    __u32 icsk_pmtu_cookie;
    const struct tcp_congestion_ops *icsk_ca_ops;
    const struct inet_connection_sock_af_ops *icsk_af_ops;
    unsigned int (*icsk_sync_mss)(struct sock *, u32);
    __u8 icsk_ca_state;
    __u8 icsk_retransmits;
    __u8 icsk_pending;
    __u8 icsk_backoff;
    __u8 icsk_syn_retries;
    __u8 icsk_probes_out;
    __u16 icsk_ext_hdr_len;
    struct {
        __u8 pending;
        __u8 quick;
        __u8 pingpong;
        __u8 blocked;
        __u32 ato;
        long unsigned int timeout;
        __u32 lrcvtime;
        __u16 last_seg_size;
        __u16 rcv_mss;
    } icsk_ack;
    struct {
        int enabled;
        int search_high;
        int search_low;
        int probe_size;
    } icsk_mtup;
    u32 icsk_ca_priv[16];
#define ICSK_CA_PRIV_SIZE (16 * sizeof(u32))
};

typedef int dma_cookie_t;

struct tcp_options_received {
    long int ts_recent_stamp;
    u32 ts_recent;
    u32 rcv_tsval;
    u32 rcv_tsecr;
    u16 saw_tstamp : 1;
    u16 tstamp_ok : 1;
    u16 dsack : 1;
    u16 wscale_ok : 1;
    u16 sack_ok : 4;
    u16 snd_wscale : 4;
    u16 rcv_wscale : 4;
    u8 eff_sacks;
    u8 num_sacks;
    u16 user_mss;
    u16 mss_clamp;
};

struct tcp_sack_block {
		u32 start_seq;
		u32 end_seq;
};

struct tcp_sock {
    struct inet_connection_sock inet_conn;
    u16 tcp_header_len;
    u16 xmit_size_goal;
    __be32 pred_flags;
    u32 rcv_nxt;
    u32 copied_seq;
    u32 rcv_wup;
    u32 snd_nxt;
    u32 snd_una;
    u32 snd_sml;
    u32 rcv_tstamp;
    u32 lsndtime;
    struct {
        struct sk_buff_head prequeue;
        struct task_struct *task;
        struct iovec *iov;
        int memory;
        int len;
        struct dma_chan *dma_chan;
        int wakeup;
        struct dma_pinned_list *pinned_list;
        dma_cookie_t dma_cookie;
    } ucopy;
    u32 snd_wl1;
    u32 snd_wnd;
    u32 max_window;
    u32 mss_cache;
    u32 window_clamp;
    u32 rcv_ssthresh;
    u32 frto_highmark;
    u16 advmss;
    u8 frto_counter;
    u8 nonagle;
    u32 srtt;
    u32 mdev;
    u32 mdev_max;
    u32 rttvar;
    u32 rtt_seq;
    u32 packets_out;
    u32 retrans_out;
    u16 urg_data;
    u8 urg_mode;
    u8 ecn_flags;
    struct tcp_options_received rx_opt;
    u32 snd_ssthresh;
    u32 snd_cwnd;
    u32 snd_cwnd_cnt;
    u32 snd_cwnd_clamp;
    u32 snd_cwnd_used;
    u32 snd_cwnd_stamp;
    u32 rcv_wnd;
    u32 write_seq;
    u32 pushed_seq;
    u32 lost_out;
    u32 sacked_out;
    u32 fackets_out;
    u32 tso_deferred;
    u32 bytes_acked;
    struct sk_buff *lost_skb_hint;
    struct sk_buff *scoreboard_skb_hint;
    struct sk_buff *retransmit_skb_hint;
    struct sk_buff *forward_skb_hint;
    struct sk_buff_head out_of_order_queue;
    struct tcp_sack_block duplicate_sack[1];
    struct tcp_sack_block selective_acks[4];
    struct tcp_sack_block recv_sack_cache[4];
    struct sk_buff *highest_sack;
    int lost_cnt_hint;
    int retransmit_cnt_hint;
    u32 lost_retrans_low;
    u8 reordering;
    u8 keepalive_probes;
    u32 prior_ssthresh;
    u32 high_seq;
    u32 retrans_stamp;
    u32 undo_marker;
    int undo_retrans;
    u32 total_retrans;
    u32 urg_seq;
    u32 snd_up;
    unsigned int keepalive_time;
    unsigned int keepalive_intvl;
    long unsigned int last_synq_overflow;
    struct {
        u32 rtt;
        u32 seq;
        u32 time;
    } rcv_rtt_est;
    struct {
        int space;
        u32 seq;
        u32 time;
    } rcvq_space;
    struct {
        u32 probe_seq_start;
        u32 probe_seq_end;
    } mtu_probe;
    struct tcp_sock_af_ops *af_specific;
    struct tcp_md5sig_info *md5sig_info;
    int linger2;
};

/* from linux/tcp.h */
enum tcp_seq_states {
        TCP_SEQ_STATE_LISTENING,
        TCP_SEQ_STATE_OPENREQ,
        TCP_SEQ_STATE_ESTABLISHED,
        TCP_SEQ_STATE_TIME_WAIT,
};

struct seq_net_private {
};

struct tcp_iter_state {
	struct seq_net_private p;
	sa_family_t family;
	enum tcp_seq_states state;
	struct sock *syn_wait_sk;
	int bucket;
	int sbucket;
	int num;
	int uid;
};

struct inet_timewait_sock {
	struct sock_common __tw_common;
#define tw_family		 __tw_common.skc_family
#define tw_state		 __tw_common.skc_state
#define tw_reuse		 __tw_common.skc_reuse
#define tw_bound_dev_if __tw_common.skc_bount_dev_if
#define tw_node			 __tw_common.skc_node
#define tw_bind_node __tw_common.skc_bind_node
#define tw_refcnt		 __tw_common.skc_refcnt
#define tw_hash			 __tw_common.skc_hash
#define tw_prot			 __tw_common.skc_prot
#define tw_net			 __tw_common.skc_net
	int tw_timeout;
	volatile unsigned char tw_substate;
	unsigned char tw_rcv_wscale;
	__be16 tw_sport;
	__be32 tw_daddr;
	__be32 tw_rcv_saddr;
	__be16 tw_dport;
	__u16 tw_num;
	__u8 tw_ipv6only : 1;
	__u16 tw_ipv6_offset;
	long unsigned int tw_ttd;
	struct inet_bind_bucket *tw_tb;
	struct hlist_node tw_death_node;
};

#define PER_CPU_RUNQUEUES (0xffffffff81676000)
#define CPU_KHZ (0xffffffff81719018)

#define FALLBACK_DEV (0xffffffff8148f120)

// from include/asm-x86_64/thread_info.h
#define TIF_SYSCALL_TRACE	0	/* syscall trace active */
#define TIF_SIGPENDING		2	/* signal pending */
#define TIF_NEED_RESCHED	3	/* rescheduling necessary */
#define TIF_SINGLESTEP		4	/* reenable singlestep on user return*/
#define TIF_IRET		5	/* force IRET */
#define TIF_SYSCALL_EMU		6	/* syscall emulation active */
#define TIF_SYSCALL_AUDIT	7	/* syscall auditing active */
#define TIF_SECCOMP		8	/* secure computing */
#define TIF_MCE_NOTIFY		10	/* notify userspace of an MCE */
#define TIF_NOTSC		16	/* TSC is not accessible in userland */
#define TIF_IA32		17	/* 32bit process */
#define TIF_FORK		18	/* ret_from_fork */
#define TIF_ABI_PENDING		19
#define TIF_MEMDIE		20
#define TIF_DEBUG		21	/* uses debug registers */
#define TIF_IO_BITMAP		22	/* uses I/O bitmap */
#define TIF_FREEZE		23	/* is freezing for suspend */
#define TIF_FORCED_TF		24	/* true if TF in eflags artificially */
#define TIF_DEBUGCTLMSR		25	/* uses thread_struct.debugctlmsr */
#define TIF_DS_AREA_MSR		26      /* uses thread_struct.ds_area_msr */
#define TIF_BTS_TRACE_TS	27      /* record scheduling event timestamps */

// from include/linux/sched.h
#define TASK_RUNNING		0
#define TASK_INTERRUPTIBLE	1
#define TASK_UNINTERRUPTIBLE	2
#define __TASK_STOPPED		4
#define __TASK_TRACED		8
/* in tsk->exit_state */
#define EXIT_ZOMBIE		16
#define EXIT_DEAD		32
/* in tsk->state again */
#define TASK_DEAD		64
#define TASK_WAKEKILL		128

/*from include/net/tcp_states.h*/
enum {
        TCP_ESTABLISHED = 1,
        TCP_SYN_SENT,
        TCP_SYN_RECV,
        TCP_FIN_WAIT1,
        TCP_FIN_WAIT2,
        TCP_TIME_WAIT,
        TCP_CLOSE,
        TCP_CLOSE_WAIT,
        TCP_LAST_ACK,
        TCP_LISTEN,
        TCP_CLOSING,    /* Now a valid state */

        TCP_MAX_STATES  /* Leave at the end! */
};

/*from include/linux/net.h*/
#define __SO_ACCEPTCON  (1 << 16)       /* performed a listen */

/*from include/linux/seq_file.h*/
#define SEQ_START_TOKEN ((void *)1)

/* original value */
#define TASK_IO_STOPPED             256
/* end of original value */

/* get_task_state() */
#define TASK_REPORT		(TASK_RUNNING | TASK_INTERRUPTIBLE | \
				 TASK_UNINTERRUPTIBLE | __TASK_STOPPED | \
				 __TASK_TRACED)

/*
 * Per process flags
 */
#define PF_ALIGNWARN	0x00000001	/* Print alignment warning msgs */
					/* Not implemented yet, only for 486*/
#define PF_STARTING	0x00000002	/* being created */
#define PF_EXITING	0x00000004	/* getting shut down */
#define PF_EXITPIDONE	0x00000008	/* pi exit done on shut down */
#define PF_VCPU		0x00000010	/* I'm a virtual CPU */
#define PF_FORKNOEXEC	0x00000040	/* forked but didn't exec */
#define PF_SUPERPRIV	0x00000100	/* used super-user privileges */
#define PF_DUMPCORE	0x00000200	/* dumped core */
#define PF_SIGNALED	0x00000400	/* killed by a signal */
#define PF_MEMALLOC	0x00000800	/* Allocating memory */
#define PF_FLUSHER	0x00001000	/* responsible for disk writeback */
#define PF_USED_MATH	0x00002000	/* if unset the fpu must be initialized before use */
#define PF_NOFREEZE	0x00008000	/* this thread should not be frozen */
#define PF_FROZEN	0x00010000	/* frozen for system suspend */
#define PF_FSTRANS	0x00020000	/* inside a filesystem transaction */
#define PF_KSWAPD	0x00040000	/* I am kswapd */
#define PF_SWAPOFF	0x00080000	/* I am in swapoff */
#define PF_LESS_THROTTLE 0x00100000	/* Throttle me less: I clean memory */
#define PF_KTHREAD	0x00200000	/* I am a kernel thread */
#define PF_RANDOMIZE	0x00400000	/* randomize virtual address space */
#define PF_SWAPWRITE	0x00800000	/* Allowed to write to swap */
#define PF_SPREAD_PAGE	0x01000000	/* Spread page cache over cpuset */
#define PF_SPREAD_SLAB	0x02000000	/* Spread some slab caches over cpuset */
#define PF_THREAD_BOUND	0x04000000	/* Thread bound to specific cpu */
#define PF_MEMPOLICY	0x10000000	/* Non-default NUMA mempolicy */
#define PF_MUTEX_TESTER	0x20000000	/* Thread belongs to the rt mutex tester */
#define PF_FREEZER_SKIP	0x40000000	/* Freezer should not count it as freezeable */
#define PF_FREEZER_NOSIG 0x80000000	/* Freezer won't send signals to it */

// sched.h
#define SIGNAL_STOP_STOPPED	0x00000001 /* job control stop in effect */
#define SIGNAL_STOP_DEQUEUED	0x00000002 /* stop signal dequeued */
#define SIGNAL_STOP_CONTINUED	0x00000004 /* SIGCONT since WCONTINUED reap */
#define SIGNAL_GROUP_EXIT	0x00000008 /* group exit in progress */

#define NGROUPS_SMALL		32
#define NGROUPS_PER_BLOCK	((unsigned int)(PAGE_SIZE / sizeof(gid_t)))

// inet_connection_sock.h
#define ICSK_TIME_RETRANS				1			/* Retransmit timer */
#define ICSK_TIME_PROBE0				3			/* Zero window probe timer */

// ptrace.h
/*
 * Ptrace flags
 *
 * The owner ship rules for task->ptrace which holds the ptrace
 * flags is simple.  When a task is running it owns it's task->ptrace
 * flags.  When the a task is stopped the ptracer owns task->ptrace.
 */

#define PT_PTRACED	0x00000001
#define PT_DTRACE	0x00000002	/* delayed trace (used on m68k, i386) */
#define PT_TRACESYSGOOD	0x00000004
#define PT_PTRACE_CAP	0x00000008	/* ptracer can follow suid-exec */
#define PT_TRACE_FORK	0x00000010
#define PT_TRACE_VFORK	0x00000020
#define PT_TRACE_CLONE	0x00000040
#define PT_TRACE_EXEC	0x00000080
#define PT_TRACE_VFORK_DONE	0x00000100
#define PT_TRACE_EXIT	0x00000200

#define PT_TRACE_MASK	0x000003f4
// asm/pgtable_64.h
#define PTRS_PER_PTE	512

// net/af_unix.h
#define UNIX_HASH_SIZE	256
#define UDP_HTABLE_SIZE 128
// capability.h
#define _LINUX_CAPABILITY_U32S_3     2
#define _KERNEL_CAPABILITY_U32S    _LINUX_CAPABILITY_U32S_3

// net/inet_hashtables.h
#define INET_LHTABLE_SIZE 32

// from debug info
#define INIT_TASK_ADDR (0xffffffff814a8340UL)
#define SOCKET_FILE_OPS (0xffffffff81376110UL)
#define UNIX_SOCKET_TABLE (0xffffffff8181b860UL)
#define UDP_HASH (0xffffffff81819880UL)
#define TCP_HASHINFO (0xffffffff81555f00UL)
#define JIFFIES ((unsigned long *)0xffffffff8156a880UL)

#define LINUX_PROC_BANNER (0xffffffff8133d090UL)

#endif /* __LINUX_TYPES_H */
