#ifndef __LINUX_TYPES_H
#define __LINUX_TYPES_H

#ifndef ____cacheline_aligned
#define ____cacheline_aligned __attribute__((__aligned__(SMP_CACHE_BYTES)))
#endif

typedef struct {
    int counter;
} atomic_t;

struct llist_node {
    struct llist_node *next;
};

struct load_weight {
    long unsigned int weight;
    long unsigned int inv_weight;
};

struct rb_node {
    long unsigned int rb_parent_color;
    struct rb_node *rb_right;
    struct rb_node *rb_left;
};

struct list_head {
    struct list_head *next;
    struct list_head *prev;
};

typedef long long unsigned int u64;

typedef long long int s64;

struct sched_statistics {
    u64 wait_start;
    u64 wait_max;
    u64 wait_count;
    u64 wait_sum;
    u64 iowait_count;
    u64 iowait_sum;
    u64 sleep_start;
    u64 sleep_max;
    s64 sum_sleep_runtime;
    u64 block_start;
    u64 block_max;
    u64 exec_max;
    u64 slice_max;
    u64 nr_migrations_cold;
    u64 nr_failed_migrations_affine;
    u64 nr_failed_migrations_running;
    u64 nr_failed_migrations_hot;
    u64 nr_forced_migrations;
    u64 nr_wakeups;
    u64 nr_wakeups_sync;
    u64 nr_wakeups_migrate;
    u64 nr_wakeups_local;
    u64 nr_wakeups_remote;
    u64 nr_wakeups_affine;
    u64 nr_wakeups_affine_attempts;
    u64 nr_wakeups_passive;
    u64 nr_wakeups_idle;
};

struct sched_entity {
    struct load_weight load;
    struct rb_node run_node;
    struct list_head group_node;
    unsigned int on_rq;
    u64 exec_start;
    u64 sum_exec_runtime;
    u64 vruntime;
    u64 prev_sum_exec_runtime;
    u64 nr_migrations;
    struct sched_statistics statistics;
    struct sched_entity *parent;
    struct cfs_rq *cfs_rq;
    struct cfs_rq *my_q;
};

struct sched_rt_entity {
    struct list_head run_list;
    long unsigned int timeout;
    unsigned int time_slice;
    int nr_cpus_allowed;
    struct sched_rt_entity *back;
    struct sched_rt_entity *parent;
    struct rt_rq *rt_rq;
    struct rt_rq *my_q;
};

struct hlist_head {
    struct hlist_node *first;
};

typedef struct cpumask {
    long unsigned int bits[1];
} cpumask_t;

struct sched_info {
    long unsigned int pcount;
    long long unsigned int run_delay;
    long long unsigned int last_arrival;
    long long unsigned int last_queued;
};

struct plist_node {
    int prio;
    struct list_head prio_list;
    struct list_head node_list;
};

struct task_rss_stat {
    int events;
    int count[3];
};

typedef int pid_t;

struct hlist_node {
    struct hlist_node *next;
    struct hlist_node **pprev;
};

struct pid_link {
    struct hlist_node node;
    struct pid *pid;
};

typedef long unsigned int cputime_t;

typedef long int __kernel_time_t;

struct g_timespec {
    __kernel_time_t tv_sec;
    long int tv_nsec;
};

struct task_cputime {
    cputime_t utime;
    cputime_t stime;
    long long unsigned int sum_exec_runtime;
};

struct sysv_sem {
    struct sem_undo_list *undo_list;
};

/* from asm/desc_defs.h */
typedef unsigned short int u16;

struct desc_struct {
        union {
                struct {
                        unsigned int a;
                        unsigned int b;
                };
                struct {
                        u16 limit0;
                        u16 base0;
                        unsigned base1: 8, type: 4, s: 1, dpl: 2, p: 1;
                        unsigned limit: 4, avl: 1, l: 1, d: 1, g: 1, base2: 8;
                };
        };
} __attribute__((packed));

struct fpu {
    union thread_xstate *state;
};

struct thread_struct {
    struct desc_struct tls_array[3];
    long unsigned int sp0;
    long unsigned int sp;
    long unsigned int usersp;
    short unsigned int es;
    short unsigned int ds;
    short unsigned int fsindex;
    short unsigned int gsindex;
    long unsigned int fs;
    long unsigned int gs;
    struct perf_event *ptrace_bps[4];
    long unsigned int debugreg6;
    long unsigned int ptrace_dr7;
    long unsigned int cr2;
    long unsigned int trap_no;
    long unsigned int error_code;
    long unsigned int has_fpu;
    struct fpu fpu;
    long unsigned int *io_bitmap_ptr;
    long unsigned int iopl;
    unsigned int io_bitmap_max;
};

typedef struct {
    long unsigned int sig[1];
} g_sigset_t;

struct sigpending {
    struct list_head list;
    g_sigset_t signal;
};

typedef long unsigned int size_t;

typedef unsigned int uid_t;

struct seccomp {
    int mode;
    struct seccomp_filter *filter;
};

typedef unsigned int u32;

typedef short unsigned int __ticketpair_t;

typedef unsigned char __ticket_t;

struct __raw_tickets {
    __ticket_t head;
    __ticket_t tail;
};

typedef struct arch_spinlock {
    union {
    	__ticketpair_t head_tail;
    	struct __raw_tickets tickets;
    };
} arch_spinlock_t;

struct raw_spinlock {
    arch_spinlock_t raw_lock;
};

typedef struct spinlock {
    union {
    	struct raw_spinlock rlock;
    };
} spinlock_t;

typedef struct raw_spinlock raw_spinlock_t;

struct plist_head {
    struct list_head node_list;
};

struct task_io_accounting {
    u64 rchar;
    u64 wchar;
    u64 syscr;
    u64 syscw;
    u64 read_bytes;
    u64 write_bytes;
    u64 cancelled_write_bytes;
};

typedef struct {
    long unsigned int bits[1];
} nodemask_t;

struct mutex {
    atomic_t count;
    spinlock_t wait_lock;
    struct list_head wait_list;
    struct task_struct *owner;
};

struct rcu_head {
    struct rcu_head *next;
    void (*func)(struct rcu_head *);
};

struct latency_record {
    long unsigned int backtrace[12];
    unsigned int count;
    long unsigned int time;
    long unsigned int max;
};

struct memcg_batch_info {
    int do_batch;
    struct mem_cgroup *memcg;
    long unsigned int nr_pages;
    long unsigned int memsw_nr_pages;
};

struct siginfo;
typedef struct siginfo siginfo_t;

struct task_struct {
    volatile long int state;
    void *stack;
    atomic_t usage;
    unsigned int flags;
    unsigned int ptrace;
    struct llist_node wake_entry;
    int on_cpu;
    int on_rq;
    int prio;
    int static_prio;
    int normal_prio;
    unsigned int rt_priority;
    const struct sched_class *sched_class;
    struct sched_entity se;
    struct sched_rt_entity rt;
    struct hlist_head preempt_notifiers;
    unsigned char fpu_counter;
    unsigned int btrace_seq;
    unsigned int policy;
    cpumask_t cpus_allowed;
    struct sched_info sched_info;
    struct list_head tasks;
    struct plist_node pushable_tasks;
    struct mm_struct *mm;
    struct mm_struct *active_mm;
    struct task_rss_stat rss_stat;
    int exit_state;
    int exit_code;
    int exit_signal;
    int pdeath_signal;
    unsigned int jobctl;
    unsigned int personality;
    unsigned int did_exec : 1;
    unsigned int in_execve : 1;
    unsigned int in_iowait : 1;
    unsigned int no_new_privs : 1;
    unsigned int sched_reset_on_fork : 1;
    unsigned int sched_contributes_to_load : 1;
    pid_t pid;
    pid_t tgid;
    long unsigned int stack_canary;
    struct task_struct *real_parent;
    struct task_struct *parent;
    struct list_head children;
    struct list_head sibling;
    struct task_struct *group_leader;
    struct list_head ptraced;
    struct list_head ptrace_entry;
    struct pid_link pids[3];
    struct list_head thread_group;
    struct completion *vfork_done;
    int *set_child_tid;
    int *clear_child_tid;
    cputime_t utime;
    cputime_t stime;
    cputime_t utimescaled;
    cputime_t stimescaled;
    cputime_t gtime;
    cputime_t prev_utime;
    cputime_t prev_stime;
    long unsigned int nvcsw;
    long unsigned int nivcsw;
    struct g_timespec start_time;
    struct g_timespec real_start_time;
    long unsigned int min_flt;
    long unsigned int maj_flt;
    struct task_cputime cputime_expires;
    struct list_head cpu_timers[3];
    const struct cred *real_cred;
    const struct cred *cred;
    struct cred *replacement_session_keyring;
    char comm[16];
    int link_count;
    int total_link_count;
    struct sysv_sem sysvsem;
    long unsigned int last_switch_count;
    struct thread_struct thread;
    struct fs_struct *fs;
    struct files_struct *files;
    struct nsproxy *nsproxy;
    struct signal_struct *signal;
    struct sighand_struct *sighand;
    g_sigset_t blocked;
    g_sigset_t real_blocked;
    g_sigset_t saved_sigmask;
    struct sigpending pending;
    long unsigned int sas_ss_sp;
    size_t sas_ss_size;
    int (*notifier)(void *);
    void *notifier_data;
    g_sigset_t *notifier_mask;
    struct audit_context *audit_context;
    uid_t loginuid;
    unsigned int sessionid;
    struct seccomp seccomp;
    u32 parent_exec_id;
    u32 self_exec_id;
    spinlock_t alloc_lock;
    struct irqaction *irqaction;
    raw_spinlock_t pi_lock;
    struct plist_head pi_waiters;
    struct rt_mutex_waiter *pi_blocked_on;
    void *journal_info;
    struct bio_list *bio_list;
    struct blk_plug *plug;
    struct reclaim_state *reclaim_state;
    struct backing_dev_info *backing_dev_info;
    struct io_context *io_context;
    long unsigned int ptrace_message;
    siginfo_t *last_siginfo;
    struct task_io_accounting ioac;
    u64 acct_rss_mem1;
    u64 acct_vm_mem1;
    cputime_t acct_timexpd;
    nodemask_t mems_allowed;
    int mems_allowed_change_disable;
    int cpuset_mem_spread_rotor;
    int cpuset_slab_spread_rotor;
    struct css_set *cgroups;
    struct list_head cg_list;
    struct robust_list_head *robust_list;
    struct compat_robust_list_head *compat_robust_list;
    struct list_head pi_state_list;
    struct futex_pi_state *pi_state_cache;
    struct perf_event_context *perf_event_ctxp[2];
    struct mutex perf_event_mutex;
    struct list_head perf_event_list;
    struct mempolicy *mempolicy;
    short int il_next;
    short int pref_node_fork;
    struct rcu_head rcu;
    struct pipe_inode_info *splice_pipe;
    struct task_delay_info *delays;
    int nr_dirtied;
    int nr_dirtied_pause;
    int latency_record_count;
    struct latency_record latency_record[32];
    long unsigned int timer_slack_ns;
    long unsigned int default_timer_slack_ns;
    struct list_head *scm_work_list;
    int curr_ret_stack;
    struct ftrace_ret_stack *ret_stack;
    long long unsigned int ftrace_timestamp;
    atomic_t trace_overrun;
    atomic_t tracing_graph_pause;
    long unsigned int trace;
    long unsigned int trace_recursion;
    struct memcg_batch_info memcg_batch;
    atomic_t ptrace_bp_refcnt;
};

struct kref {
    atomic_t refcount;
};

struct new_utsname {
    char sysname[65];
    char nodename[65];
    char release[65];
    char version[65];
    char machine[65];
    char domainname[65];
};

struct uts_namespace {
    struct kref kref;
    struct new_utsname name;
    struct user_namespace *user_ns;
};

struct pidmap {
    atomic_t nr_free;
    void *page;
};

typedef unsigned int gid_t;

struct pid_namespace {
    struct kref kref;
    struct pidmap pidmap[128];
    int last_pid;
    struct task_struct *child_reaper;
    struct kmem_cache *pid_cachep;
    unsigned int level;
    struct pid_namespace *parent;
    struct vfsmount *proc_mnt;
    struct bsd_acct_struct *bacct;
    gid_t pid_gid;
    int hide_pid;
    int reboot;
};

struct nsproxy {
    atomic_t count;
    struct uts_namespace *uts_ns;
    struct ipc_namespace *ipc_ns;
    struct mnt_namespace *mnt_ns;
    struct pid_namespace *pid_ns;
    struct net *net_ns;
};

typedef long int clock_t;

typedef unsigned int dev_t;

struct winsize {
    short unsigned int ws_row;
    short unsigned int ws_col;
    short unsigned int ws_xpixel;
    short unsigned int ws_ypixel;
};

typedef struct {
    long int counter;
} atomic_long_t;

typedef void (*work_func_t)(struct work_struct *work);

struct work_struct {
    atomic_long_t data;
    struct list_head entry;
    work_func_t func;
};

struct tty_bufhead {
    struct work_struct work;
    spinlock_t lock;
    struct tty_buffer *head;
    struct tty_buffer *tail;
    struct tty_buffer *free;
    int memory_used;
};

typedef struct __wait_queue_head {
    spinlock_t lock;
    struct list_head task_list;
} wait_queue_head_t;

struct tty_struct {
    int magic;
    struct kref kref;
    struct device *dev;
    struct tty_driver *driver;
    const struct tty_operations *ops;
    int index;
    struct mutex ldisc_mutex;
    struct tty_ldisc *ldisc;
    struct mutex termios_mutex;
    spinlock_t ctrl_lock;
    struct ktermios *termios;
    struct ktermios *termios_locked;
    struct termiox *termiox;
    char name[64];
    struct pid *pgrp;
    struct pid *session;
    long unsigned int flags;
    int count;
    struct winsize winsize;
    unsigned char stopped : 1;
    unsigned char hw_stopped : 1;
    unsigned char flow_stopped : 1;
    unsigned char packet : 1;
    unsigned char low_latency : 1;
    unsigned char warned : 1;
    unsigned char ctrl_status;
    unsigned int receive_room;
    struct tty_struct *link;
    struct fasync_struct *fasync;
    struct tty_bufhead buf;
    int alt_speed;
    wait_queue_head_t write_wait;
    wait_queue_head_t read_wait;
    struct work_struct hangup_work;
    void *disc_data;
    void *driver_data;
    struct list_head tty_files;
    unsigned int column;
    unsigned char lnext : 1;
    unsigned char erasing : 1;
    unsigned char raw : 1;
    unsigned char real_raw : 1;
    unsigned char icanon : 1;
    unsigned char closing : 1;
    unsigned char echo_overrun : 1;
    short unsigned int minimum_to_wake;
    long unsigned int overrun_time;
    int num_overrun;
    long unsigned int process_char_map[4];
    char *read_buf;
    int read_head;
    int read_tail;
    int read_cnt;
    long unsigned int read_flags[64];
    unsigned char *echo_buf;
    unsigned int echo_pos;
    unsigned int echo_cnt;
    int canon_data;
    long unsigned int canon_head;
    unsigned int canon_column;
    struct mutex atomic_read_lock;
    struct mutex atomic_write_lock;
    struct mutex output_lock;
    struct mutex echo_lock;
    unsigned char *write_buf;
    int write_cnt;
    spinlock_t read_lock;
    struct work_struct SAK_work;
    struct tty_port *port;
};

struct kobject {
    const char *name;
    struct list_head entry;
    struct kobject *parent;
    struct kset *kset;
    struct kobj_type *ktype;
    struct sysfs_dirent *sd;
    struct kref kref;
    unsigned int state_initialized : 1;
    unsigned int state_in_sysfs : 1;
    unsigned int state_add_uevent_sent : 1;
    unsigned int state_remove_uevent_sent : 1;
    unsigned int uevent_suppress : 1;
};

struct cdev {
    struct kobject kobj;
    struct module *owner;
    const struct file_operations *ops;
    struct list_head list;
    dev_t dev;
    unsigned int count;
};

typedef unsigned int tcflag_t;

typedef unsigned char cc_t;

typedef unsigned int speed_t;

struct ktermios {
    tcflag_t c_iflag;
    tcflag_t c_oflag;
    tcflag_t c_cflag;
    tcflag_t c_lflag;
    cc_t c_line;
    cc_t c_cc[19];
    speed_t c_ispeed;
    speed_t c_ospeed;
};

struct tty_driver {
    int magic;
    struct kref kref;
    struct cdev cdev;
    struct module *owner;
    const char *driver_name;
    const char *name;
    int name_base;
    int major;
    int minor_start;
    int minor_num;
    int num;
    short int type;
    short int subtype;
    struct ktermios init_termios;
    int flags;
    struct proc_dir_entry *proc_entry;
    struct tty_driver *other;
    struct tty_struct **ttys;
    struct ktermios **termios;
    struct ktermios **termios_locked;
    void *driver_state;
    const struct tty_operations *ops;
    struct list_head tty_drivers;
};

typedef long unsigned int pteval_t;

typedef struct {
    pteval_t pte;
};

enum pid_type {PIDTYPE_PID, PIDTYPE_PGID, PIDTYPE_SID, PIDTYPE_MAX};

struct fdtable {
    unsigned int max_fds;
    struct file **fd;
    fd_set *close_on_exec;
    fd_set *open_fds;
    struct rcu_head rcu;
    struct fdtable *next;
};

struct embedded_fd_set {
    long unsigned int fds_bits[1];
};

struct files_struct {
    atomic_t count;
    struct fdtable *fdt;
    struct fdtable fdtab;
    spinlock_t file_lock;
    int next_fd;
    struct embedded_fd_set close_on_exec_init;
    struct embedded_fd_set open_fds_init;
    struct file *fd_array[64];
};

typedef void __signalfn_t(int);
typedef __signalfn_t __user *__sighandler_t;

typedef void __restorefn_t(void);
typedef __restorefn_t __user *__sigrestore_t;

typedef struct {
    long unsigned int sig[1];
} sigset_t;

struct sigaction {
    __sighandler_t sa_handler;
    long unsigned int sa_flags;
    __sigrestore_t sa_restorer;
    sigset_t sa_mask;
};

struct k_sigaction {
    struct sigaction sa;
};

struct sighand_struct {
    atomic_t count;
    struct k_sigaction action[64];
    spinlock_t siglock;
    wait_queue_head_t signalfd_wqh;
};

typedef union ktime {
    s64 tv64;
} ktime_t;

struct timerqueue_node {
    struct rb_node node;
    ktime_t expires;
};

struct hrtimer {
    struct timerqueue_node node;
    ktime_t _softexpires;
    enum hrtimer_restart (*function)(struct hrtimer *);
    struct hrtimer_clock_base *base;
    long unsigned int state;
    int start_pid;
    void *start_site;
    char start_comm[16];
};

struct cpu_itimer {
    cputime_t expires;
    cputime_t incr;
    u32 error;
    u32 incr_error;
};

struct thread_group_cputimer {
    struct task_cputime cputime;
    int running;
    raw_spinlock_t lock;
};

struct rlimit {
    long unsigned int rlim_cur;
    long unsigned int rlim_max;
};

struct pacct_struct {
    int ac_flag;
    long int ac_exitcode;
    long unsigned int ac_mem;
    cputime_t ac_utime;
    cputime_t ac_stime;
    long unsigned int ac_minflt;
    long unsigned int ac_majflt;
};

struct rw_semaphore {
    long int count;
    raw_spinlock_t wait_lock;
    struct list_head wait_list;
};

struct signal_struct {
    atomic_t sigcnt;
    atomic_t live;
    int nr_threads;
    wait_queue_head_t wait_chldexit;
    struct task_struct *curr_target;
    struct sigpending shared_pending;
    int group_exit_code;
    int notify_count;
    struct task_struct *group_exit_task;
    int group_stop_count;
    unsigned int flags;
    struct list_head posix_timers;
    struct hrtimer real_timer;
    struct pid *leader_pid;
    ktime_t it_real_incr;
    struct cpu_itimer it[2];
    struct thread_group_cputimer cputimer;
    struct task_cputime cputime_expires;
    struct list_head cpu_timers[3];
    struct pid *tty_old_pgrp;
    int leader;
    struct tty_struct *tty;
    struct autogroup *autogroup;
    cputime_t utime;
    cputime_t stime;
    cputime_t cutime;
    cputime_t cstime;
    cputime_t gtime;
    cputime_t cgtime;
    cputime_t prev_utime;
    cputime_t prev_stime;
    long unsigned int nvcsw;
    long unsigned int nivcsw;
    long unsigned int cnvcsw;
    long unsigned int cnivcsw;
    long unsigned int min_flt;
    long unsigned int maj_flt;
    long unsigned int cmin_flt;
    long unsigned int cmaj_flt;
    long unsigned int inblock;
    long unsigned int oublock;
    long unsigned int cinblock;
    long unsigned int coublock;
    long unsigned int maxrss;
    long unsigned int cmaxrss;
    struct task_io_accounting ioac;
    long long unsigned int sum_sched_runtime;
    struct rlimit rlim[16];
    struct pacct_struct pacct;
    struct taskstats *stats;
    unsigned int audit_tty;
    struct tty_audit_buf *tty_audit_buf;
    struct rw_semaphore threadgroup_fork_lock;
    int oom_adj;
    int oom_score_adj;
    int oom_score_adj_min;
    struct mutex cred_guard_mutex;
};

//from debug info
#define    INIT_TASK_ADDR    0xffffffff81c0d020UL

#endif /* __LINUX_TYPES_H */
