#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "linux_types.h"
#include "addr.h"
#include "net.h"
#include "array.h"


int timer_pending(const struct timer_list *timer)
{
    return timer->entry.next != NULL;
}

int hlist_empty(struct hlist_head *h)
{
    return !h->first;
}

struct hlist_node *next_hlist_node(struct hlist_node *node, struct sched_struct *handle)
{
    struct hlist_node **usr_next_node = HREF(node, next);

    return *usr_next_node;
}

struct unix_sock *unix_sk(struct sock *sk, struct sched_struct *handle)
{
    struct unix_sock *usk = (struct unix_sock *)sk;
    struct unix_sock *p = ADDRX(usk);
    wait_queue_head_t *tmp_peer_wait = HREF(usk, peer_wait);

    return p;
}

struct inet_timewait_sock *tw_head(struct hlist_head *head)
{
	return hlist_empty(head) ? NULL :	list_entry(head->first, struct inet_timewait_sock, tw_node);
}

struct inet_timewait_sock *tw_next(struct inet_timewait_sock *tw)
{
	return tw->tw_node.next ?
			hlist_entry(tw->tw_node.next, typeof(*tw), tw_node) : NULL;
}

struct tcp_sock *tcp_sk(struct sock *sk, struct sched_struct *handle)
{
    struct tcp_sock *tsk = (struct tcp_sock *)sk;
    struct tcp_sock *p = ADDRX(tsk);
    int *linger2 = HREF(tsk, linger2);

    return p;
}

struct inet_sock *inet_sk(struct sock *sk, struct sched_struct *handle)
{
    struct inet_sock *isk = (struct inet_sock *)sk;
    struct inet_sock *p = ADDRX(isk);
    struct ip_mc_socklist **tmp_mc_list = HREF(isk, mc_list);

    return p;
}

struct inet_connection_sock *inet_csk(struct sock *sk, struct sched_struct *handle)
{
    struct inet_connection_sock *icsk = (struct inet_connection_sock *)sk;
    struct inet_connection_sock *p = ADDRX(icsk);
    u32 *tmp_icsk_ca_priv = HREF(icsk, icsk_ca_priv[15]);

    return p;
}

struct sock *__sk_head(struct hlist_head *head)
{
	return hlist_entry(head->first, struct sock, sk_node);
}

struct sock *sk_head(struct hlist_head *head)
{
	return hlist_empty(head) ? NULL : __sk_head(head);
}

int UNIX_ABSTRACT(struct unix_sock *u, struct sched_struct *handle)
{
    unsigned int *user_hash = HREF(u->addr, hash);

    return (*user_hash != UNIX_HASH_SIZE);
}

struct inode *SOCK_INODE(struct socket *socket, struct sched_struct *handle)
{
    struct socket_alloc *s_alloc = container_of(socket, struct socket_alloc, socket);
    struct inode *vfs_inode = HREF(s_alloc, vfs_inode);

    return vfs_inode;
}

unsigned long sock_i_uid(struct sock *sk, struct sched_struct *handle)
{
    int uid;
    struct inode *inode = SOCK_INODE(sk->sk_socket, handle);

    uid = sk->sk_socket ? inode->i_uid : 0;
    UNMAP(inode);

    return uid;
}

unsigned long sock_i_ino(struct sock *sk, struct sched_struct *handle)
{
    unsigned long ino;
    struct inode *inode = SOCK_INODE(sk->sk_socket, handle);

    ino = sk->sk_socket ? inode->i_ino : 0;
    UNMAP(inode);

    return ino;
}

struct sock *sk_next(struct sock *sk)
{
    return sk->sk_node.next ?
        hlist_entry(sk->sk_node.next, struct sock, sk_node) : NULL;
}	

struct sock *next_socket(struct sock *sk, struct sched_struct *handle)
{
    struct sock *sock = ADDRX(sk);
    struct sock *next_sock = sk_next(sock);  	

    UNMAP(sock);

    return next_sock;
}

struct sock *next_sock_from_node(struct sock *sk, struct hlist_node *node, struct sched_struct *handle)
{
    struct sock *res = NULL;

    if(node->next != NULL){
        struct sock *next_sock = hlist_entry(node->next, typeof(*sk), sk_node);
        struct sock *usr_next_sock = ADDRX(next_sock);	
        res = usr_next_sock;
        UNMAP(sk);
    }

    return res;
}

int reqsk_queue_len(struct request_sock_queue *queue, struct sched_struct *handle)
{
    int res = 0;

    if(queue->listen_opt != NULL){
        int *usr_qlen = HREF(queue->listen_opt, qlen);

        res = *usr_qlen;
        UNMAP(usr_qlen);
    }

    return res;
}	

void init_tcp_hashlist(struct sched_struct *handle)
{
    struct inet_hashinfo *hashinfo = (struct inet_hashinfo *)TCP_HASHINFO;

    tcp_hashinfo = ADDRX(hashinfo);
    ehash_list = ADDRX(tcp_hashinfo->ehash);
}

void tcp_hashlist_nextpage(struct sched_struct *handle)
{
    UNMAP(ehash_list);

    struct inet_ehash_bucket *next = (struct inet_ehash_bucket *)((unsigned long)tcp_hashinfo->ehash + (unsigned long)(PAGE_SIZE * ehash_cnt));

    ehash_list = ADDRX(next);
}

void close_tcp_hashlist(struct sched_struct *handle)
{
    UNMAP(tcp_hashinfo);
    UNMAP(ehash_list);
    ehash_cnt = 0;
}

void *listening_get_next(struct seq_file *seq, void *sock_addr, struct sched_struct *handle)
{
    struct inet_connection_sock *icsk;
    struct hlist_node *node;
    struct sock *sk = sock_addr;
    struct tcp_iter_state* st = seq->private;
//        struct net *net = seq_file_net(seq);

    if (!sk) {
        st->bucket = 0;
        sk = sk_head(&tcp_hashinfo->listening_hash[0]);
        goto get_sk;
    }

    ++st->num;

    if (st->state == TCP_SEQ_STATE_OPENREQ) {
        struct request_sock *req = ADDRX(sock_addr);
        
        icsk = inet_csk(st->syn_wait_sk, handle);
        struct request_sock *del = req;
        req = ADDRX(req->dl_next);
        UNMAP(del);

        while (1) {
            while (req) {
                int *usr_family = HREF((struct request_sock_ops *)req->rsk_ops,
                                       family);
                
                if (*usr_family == st->family) {
                    UNMAP(usr_family);
                    sock_addr = req;
                    goto out;
                }
                UNMAP(usr_family);
                del = req;
                req = ADDRX(req->dl_next);
                UNMAP(del);
            }

            u32 *usr_entries = HREF(icsk->icsk_accept_queue.listen_opt,
                                    nr_table_entries);

            if (++st->sbucket >= *usr_entries){
                UNMAP(usr_entries);
                break;
            }
get_req:
            del = req;

            struct request_sock **usr_syn_table = HREF(icsk->icsk_accept_queue.listen_opt, syn_table[st->sbucket]);
                                                       
            req = ADDRX(*usr_syn_table);
            UNMAP(del);
            UNMAP(usr_syn_table);
        }
        sk = next_socket(st->syn_wait_sk, handle);
							
        st->state = TCP_SEQ_STATE_LISTENING;
//                read_unlock_bh(&icsk->icsk_accept_queue.syn_wait_lock);
    } else {
        icsk = inet_csk(sk, handle);
//                read_lock_bh(&icsk->icsk_accept_queue.syn_wait_lock);
        if (reqsk_queue_len(&icsk->icsk_accept_queue, handle)){
            UNMAP(icsk);
            goto start_req;
								}
//                read_unlock_bh(&icsk->icsk_accept_queue.syn_wait_lock);
        sk = next_socket(sk, handle);
    }
get_sk:
    my_sk_for_each_from(sk, node, handle) {
//                if (socket->sk_family == st->family && net_eq(sock_net(sk), net)) {
        struct sock *socket = ADDRX(sk);
        if (socket->sk_family == st->family ) {
            sock_addr = sk;
            UNMAP(socket);
            goto out;
        }
        icsk = inet_csk(sk, handle);
//                read_lock_bh(&icsk->icsk_accept_queue.syn_wait_lock);
        if (reqsk_queue_len(&icsk->icsk_accept_queue, handle)) {
start_req:
            st->uid         = sock_i_uid(socket, handle);
            st->syn_wait_sk = sk;
            st->state       = TCP_SEQ_STATE_OPENREQ;
            st->sbucket     = 0;
            goto get_req;
        }
//               read_unlock_bh(&icsk->icsk_accept_queue.syn_wait_lock);
				}
    if (++st->bucket < INET_LHTABLE_SIZE) {
        sk = sk_head(&tcp_hashinfo->listening_hash[st->bucket]);
        goto get_sk;
    }
    sock_addr = NULL;
out:
    return sock_addr;
}

void *listening_get_idx(struct seq_file *seq, loff_t *pos,
                        struct sched_struct *handle)
{
    void *rc = listening_get_next(seq, NULL, handle);

    while (rc && *pos) {
        rc = listening_get_next(seq, rc, handle);
        --*pos;
    }
    return rc;
}

void *established_get_first(struct seq_file *seq, struct sched_struct *handle)
{
    struct tcp_iter_state* st = seq->private;
//        struct net *net = seq_file_net(seq);
    void *rc = NULL;

    for (ehash_cnt = 0; ehash_cnt < tcp_hashinfo->ehash_size / (PAGE_SIZE / sizeof(struct inet_ehash_bucket)); ehash_cnt++) {
        tcp_hashlist_nextpage(handle);
        for(st->bucket = 0; st->bucket < (PAGE_SIZE / sizeof(struct inet_ehash_bucket)); st->bucket++) {
            struct sock *sk;
            struct hlist_node *node;
            struct inet_timewait_sock *tw;
//                rwlock_t *lock = inet_ehash_lockp(&tcp_hashinfo, st->bucket);

//                read_lock_bh(lock);
            my_sk_for_each(sk, node, &ehash_list[st->bucket].chain, handle) {
                struct sock *socket = ADDRX(sk);
//                        if (sk->sk_family != st->family || !net_eq(sock_net(sk), net)) {
                if(socket->sk_family != st->family){
                    UNMAP(socket);
                    continue;
                }
                UNMAP(socket);
                rc = sk;
                goto out;
            }
            st->state = TCP_SEQ_STATE_TIME_WAIT;
            my_inet_twsk_for_each(tw, node,
                                  &ehash_list[st->bucket].twchain, handle) {
                struct inet_timewait_sock *socket = ADDRX(tw);
//                        if (tw->tw_family != st->family || !net_eq(twsk_net(tw), net)) {
                if(socket->tw_family != st->family){
                    UNMAP(socket);
                    continue;
                }
                UNMAP(socket);
                rc = tw;
                goto out;
            }
//              read_unlock_bh(lock);
            st->state = TCP_SEQ_STATE_ESTABLISHED;
        }
    }

out:
    return rc;
}

void *established_get_next(struct seq_file *seq, void *cur, struct sched_struct *handle)
{
    struct sock *sk = cur;
    struct inet_timewait_sock *tw;
    struct inet_timewait_sock *tw_addr;
    struct hlist_node *node;
    struct tcp_iter_state* st = seq->private;
//        struct net *net = seq_file_net(seq);

    ++st->num;

    if (st->state == TCP_SEQ_STATE_TIME_WAIT) {
        struct inet_timewait_sock *old_tw = ADDRX(cur);
        tw_addr = tw_next(old_tw);
        UNMAP(old_tw);
        if(tw_addr == NULL)
            goto null_tw;
        tw = ADDRX(tw_addr);
get_tw:
//                while (tw && (tw->tw_family != st->family || !net_eq(twsk_net(tw), net))) {
        while(tw && tw->tw_family != st->family){
            struct inet_timewait_sock *del = tw;
            tw_addr = tw_next(tw);
            if(tw_addr == NULL)
                goto null_tw;
            tw = ADDRX(tw_addr);
            UNMAP(del);
        }
null_tw:
        if (tw_addr) {
            cur = tw_addr;
            goto out;
        }
//                read_unlock_bh(inet_ehash_lockp(&tcp_hashinfo, st->bucket));
        st->state = TCP_SEQ_STATE_ESTABLISHED;
        if(++st->bucket < (PAGE_SIZE / sizeof(struct inet_ehash_bucket))){
//                        read_lock_bh(inet_ehash_lockp(&tcp_hashinfo, st->bucket));
            sk = sk_head(&ehash_list[st->bucket].chain);
        } else if(++ehash_cnt < tcp_hashinfo->ehash_size / (PAGE_SIZE / sizeof(struct inet_ehash_bucket))) {
            st->bucket = 0;
            tcp_hashlist_nextpage(handle);
            sk = sk_head(&ehash_list[0].chain);
        } else {
            cur = NULL;
            goto out;
        }
    } else 
        sk = next_socket(sk, handle);

    my_sk_for_each_from(sk, node, handle) {
        struct sock *socket = ADDRX(sk);
//                if (sk->sk_family == st->family && net_eq(sock_net(sk), net))
        if(socket->sk_family == st->family){
            UNMAP(socket);
            goto found;
        }
        UNMAP(socket);
    }
    
    st->state = TCP_SEQ_STATE_TIME_WAIT;
    tw_addr = tw_head(&ehash_list[st->bucket].twchain);
    if(tw_addr == NULL)
        goto null_tw;

    tw = ADDRX(tw_addr);
    goto get_tw;

found:
    cur = sk;
out:
    return cur;
}

static void *established_get_idx(struct seq_file *seq, loff_t pos, struct sched_struct *handle)
{
    void *rc = established_get_first(seq, handle);

    while (rc && pos) {
        rc = established_get_next(seq, rc, handle);
        --pos;
    }
    return rc;
}

static void *tcp_get_idx(struct seq_file *seq, loff_t pos, struct sched_struct *handle)
{
    void *rc;
    struct tcp_iter_state* st = seq->private;

//        inet_listen_lock(&tcp_hashinfo);
    st->state = TCP_SEQ_STATE_LISTENING;
    rc        = listening_get_idx(seq, &pos, handle);

    if (!rc) {
//                inet_listen_unlock(&tcp_hashinfo);
        st->state = TCP_SEQ_STATE_ESTABLISHED;
        rc        = established_get_idx(seq, pos, handle);
    }

    return rc;
}

static void *tcp_seq_start(struct seq_file *seq, loff_t *pos, struct sched_struct *handle)
{
    struct tcp_iter_state* st = seq->private;
    st->state = TCP_SEQ_STATE_LISTENING;
    st->num = 0;
    return *pos ? tcp_get_idx(seq, *pos - 1, handle) : SEQ_START_TOKEN;
}

static void *tcp_seq_next(struct seq_file *seq, void *v, loff_t *pos, struct sched_struct *handle)
{
    void *rc = NULL;
    struct tcp_iter_state* st;
#ifdef DEBUG
    fprintf(stderr, "tcp_seq_next_start\n");
#endif
    if (v == SEQ_START_TOKEN) {
        rc = tcp_get_idx(seq, 0, handle);
        goto out;
    }
    st = seq->private;
    
    switch (st->state) {
    case TCP_SEQ_STATE_OPENREQ:
    case TCP_SEQ_STATE_LISTENING:
#ifdef DEBUG
        fprintf(stderr, "listening_get\n");
#endif
        rc = listening_get_next(seq, v, handle);
        if (!rc) {
//                        inet_listen_unlock(&tcp_hashinfo);
            st->state = TCP_SEQ_STATE_ESTABLISHED;
#ifdef DEBUG
            fprintf(stderr, "listen->established_get_first\n");
#endif
            rc        = established_get_first(seq, handle);
        }
        break;
    case TCP_SEQ_STATE_ESTABLISHED:
    case TCP_SEQ_STATE_TIME_WAIT:
#ifdef DEBUG
        fprintf(stderr, "establish\n");
#endif
        rc = established_get_next(seq, v, handle);
        break;
    }
out:
    ++*pos;
    return rc;
}


int unix_format(char *p, void *sock_addr, void *v, struct sched_struct *handle)
{
    char str[256];

    if(!strcmp(p, "")){
        sprintf(str, "Num       RefCount Protocol Flags    Type St "
                "Inode Path\n");
        strcpy(p, str);
    } else {
        sprintf(str, "%p", sock_addr);
        if(strstr(p, str)) return 1;
    }

    struct sock *s = v;
    struct unix_sock *u = unix_sk(sock_addr, handle);
    
    sprintf(str, "%p: %08X %08X %08X %04X %02X %5lu",
            sock_addr,
            atomic_read(s->sk_refcnt),
            0,
            s->sk_state == TCP_LISTEN ? __SO_ACCEPTCON : 0,
            s->sk_type,
            s->sk_socket ?
            (s->sk_state == TCP_ESTABLISHED ? SS_CONNECTED : SS_UNCONNECTED) :
            (s->sk_state == TCP_ESTABLISHED ? SS_CONNECTING : SS_DISCONNECTING),
            sock_i_ino(s, handle));
    
    strcat(p, str);
    
    if(u->addr) {
        int i, len;
        sprintf(str, " ");
        strcat(p, str);

        i = 0;
        int *user_addr_len = HREF(u->addr, len);

        len = *user_addr_len - sizeof(short);

        if(!(UNIX_ABSTRACT(u, handle))){
            len--;
        } else {
            sprintf(str, "@");
            strcat(p, str);
            i++;
        }

        struct sockaddr_un *addr_name = GREF(u->addr, name);
        struct sockaddr_un *user_addr_name = ADDRX(addr_name);

        for( ; i < len; i++){
            sprintf(str, "%c", user_addr_name->sun_path[i]);
            strcat(p, str);
        }
    }	

    sprintf(str, "\n");
    strcat(p, str);

    return 0;
}

void udp_format(char *p, struct sock *sock_addr, struct sock *udp_sock, struct sched_struct *handle, int *cnt, int *len)
{
    char str[256];
    if(!strcmp(p, "")){
        sprintf(str, "  sl  local_address rem_address   st tx_queue "
                "rx_queue tr tm->when retrnsmt   uid  timeout "
                "inode ref pointer drops\n");
        strcpy(p, str);
    }


    struct inet_sock *inet = inet_sk(sock_addr, handle);
//	struct inet_sock *inet = inet_sk(udp_sock);
    __be32 dest = inet->daddr;
    __be32 src	= inet->rcv_saddr;
    __u16 destp	= ntohs(inet->dport);
    __u16 srcp	= ntohs(inet->sport);

    sprintf(str, "%4d: %08X:%04X %08X:%04X"
            " %02X %08X:%08X %02X:%08lX %08X %5lu %8d %lu %d %p %d%n\n",
            *cnt, src, srcp, dest, destp, udp_sock->sk_state,
            atomic_read(udp_sock->sk_wmem_alloc),
            atomic_read(udp_sock->sk_rmem_alloc),
            0, 0L, 0, sock_i_uid(udp_sock, handle), 0, sock_i_ino(udp_sock, handle),
            atomic_read(udp_sock->sk_refcnt), sock_addr,
            atomic_read(udp_sock->sk_drops), len);
    strcat(p, str);	
}

void get_tcp4_sock(char *p, struct sock *sock_addr, struct sched_struct *handle, int i, int *len)
{
    char str[256];
    int timer_active;
    unsigned long timer_expires;
    struct sock *sk = ADDRX(sock_addr);
    struct tcp_sock *tp = tcp_sk(sock_addr, handle);
    struct inet_connection_sock *icsk = inet_csk(sock_addr, handle);
    struct inet_sock *inet = inet_sk(sock_addr, handle);
    __be32 dest = inet->daddr;
    __be32 src = inet->rcv_saddr;
    __u16 destp = ntohs(inet->dport);
    __u16 srcp = ntohs(inet->sport);
    unsigned long *jiffies = ADDRX(JIFFIES);

    if(!strcmp(p, "")){
        sprintf(str,
                "  sl  local_address rem_address   st tx_queue "
                "rx_queue tr tm->when retrnsmt   uid  timeout "
                "inode\n");
        strcpy(p, str);
    }

    if (icsk->icsk_pending == ICSK_TIME_RETRANS) {
        timer_active    = 1;
        timer_expires   = icsk->icsk_timeout;
    } else if (icsk->icsk_pending == ICSK_TIME_PROBE0) {
        timer_active    = 4;
        timer_expires   = icsk->icsk_timeout;
    } else if (timer_pending(&sk->sk_timer)) {
        timer_active    = 2;
        timer_expires   = sk->sk_timer.expires;
    } else {
        timer_active    = 0;
        timer_expires = *jiffies;
    }

    sprintf(str, "%4d: %08X:%04X %08X:%04X %02X %08X:%08X %02X:%08lX "
            "%08X %5lu %8d %lu %d %p %ld %ld %u %u %d%n",
            i, src, srcp, dest, destp, sk->sk_state,
            tp->write_seq - tp->snd_una,
            sk->sk_state == TCP_LISTEN ? sk->sk_ack_backlog :
            (tp->rcv_nxt - tp->copied_seq),
            timer_active,
            jiffies_to_clock_t(timer_expires - *jiffies),
            icsk->icsk_retransmits,
            sock_i_uid(sk, handle),
            icsk->icsk_probes_out,
            sock_i_ino(sk, handle),
            atomic_read(sk->sk_refcnt), sock_addr,
            jiffies_to_clock_t(icsk->icsk_rto),
            jiffies_to_clock_t(icsk->icsk_ack.ato),
            (icsk->icsk_ack.quick << 1) | icsk->icsk_ack.pingpong,
            tp->snd_cwnd,
            tp->snd_ssthresh >= 0xFFFF ? -1 : tp->snd_ssthresh,
            len);
    strcat(p, str);

    UNMAP(sk);	
    UNMAP(tp);
    UNMAP(icsk);
    UNMAP(inet);
    UNMAP(jiffies);
}

void get_timewait4_sock(char *p, void *sock_addr, struct sched_struct *handle, int i, int *len)
{
    char str[256];
    __be32 dest, src;
    __u16 destp, srcp;
    struct inet_timewait_sock *tw = ADDRX(sock_addr);
    unsigned long *jiffies = ADDRX(JIFFIES);

    int ttd = tw->tw_ttd - *jiffies;

    if (ttd < 0)
        ttd = 0;

    dest  = tw->tw_daddr;
    src   = tw->tw_rcv_saddr;
    destp = ntohs(tw->tw_dport);
    srcp  = ntohs(tw->tw_sport);

    sprintf(str, "%4d: %08X:%04X %08X:%04X"
            " %02X %08X:%08X %02X:%08lX %08X %5d %8d %d %d %p",
            i, src, srcp, dest, destp, tw->tw_substate, 0, 0,
            3, jiffies_to_clock_t(ttd), 0, 0, 0, 0,
            atomic_read(tw->tw_refcnt), sock_addr, len);
    
    strcat(p, str);
    UNMAP(tw);
    UNMAP(jiffies);
}

int unix_show(char *p, struct sock *sk, struct sched_struct *handle)
{
    struct sock *unix_sock = ADDRX(sk);
    __u32 *user_tmp = HREF(sk, sk_mark);

    unix_format(p, sk, unix_sock, handle);

    struct sock *next_sock = next_socket(sk, handle);

    while(next_sock != NULL){
        unix_sock = ADDRX(next_sock);
        user_tmp = HREF(next_sock, sk_mark);

        unix_format(p, next_sock, unix_sock, handle);
        next_sock = next_socket(next_sock, handle);
    }

    return 1;
}

int udp_show(char *p, struct sock *sk, struct sched_struct *handle, int *i)
{
    struct sock *udp_sock = ADDRX(sk);
    __u32 *user_tmp = HREF(sk, sk_mark);

    if(udp_sock->sk_family == 2){
        int len;
        udp_format(p, sk, udp_sock, handle, i, &len);
    }

    struct sock *next_sock = next_socket(sk, handle);

    while(next_sock != NULL){
        udp_sock = ADDRX(next_sock);
        user_tmp = HREF(next_sock, sk_mark);

        if(udp_sock->sk_family == 2){
            int len;
            udp_format(p, next_sock, udp_sock, handle, i, &len);
        }
        next_sock = next_socket(next_sock, handle);
    }

    return 1;
}

#define TMPSZ 150

int tcp4_seq_show(struct seq_file *seq, void *v, char *p, struct sched_struct *handle)
{
    struct tcp_iter_state* st;
    int len;
    char str[256];

    if (v == SEQ_START_TOKEN) {
        sprintf(str, "%-*s\n", TMPSZ - 1,
                "  sl  local_address rem_address   st tx_queue "
                "rx_queue tr tm->when retrnsmt   uid  timeout "
                "inode");
        strcpy(p, str);
        goto out;
    }
    st = seq->private;

    switch (st->state) {
    case TCP_SEQ_STATE_LISTENING:
    case TCP_SEQ_STATE_ESTABLISHED:
#ifdef DEBUG
        fprintf(stderr, "get_tcp4_sock\n");
#endif
        get_tcp4_sock(p, v, handle, st->num, &len);
        break;

    case TCP_SEQ_STATE_OPENREQ:
#ifdef DEBUG
        fprintf(stderr, "get_openreq4\n");
#endif
//                get_openreq4(st->syn_wait_sk, v, seq, st->num, st->uid, &len);
        break;

    case TCP_SEQ_STATE_TIME_WAIT:
#ifdef DEBUG
        fprintf(stderr, "get_timewait4_sock\n");
#endif
        get_timewait4_sock(p, v, handle, st->num, &len);
        break;
    }
//        sprintf(str, "%*s\n", TMPSZ - 1 - len, "");
    sprintf(str, "\n");
    strcat(p, str);

out:
    return 0;
}

int tcp_open(char *p, struct sched_struct *handle)
{
    struct seq_file seq;
    loff_t pos = 0;
    void *v = NULL;

    seq.private = (void *)malloc(sizeof(struct tcp_iter_state));
//	seq_start(&seq, &pos);	
    memset(seq.private, 0, sizeof(struct tcp_iter_state));
    ((struct tcp_iter_state *)seq.private)->family = 2;

    init_tcp_hashlist(handle);

    v = tcp_seq_start(&seq, &pos, handle);
    while(v != NULL){	
        tcp4_seq_show(&seq, v, p, handle);
        v = tcp_seq_next(&seq, v, &pos, handle);	
    }
    close_tcp_hashlist(handle);

    free(seq.private);
    return 0;
}
