#ifndef __UNIX_H
#define __UNIX_H

#include "linux_types.h"
#include "sched.h"

#define container_of(ptr, type, member) ({                \
        const typeof( ((type *)0)->member ) *__mptr = (ptr);    \
        (type *)( (char *)__mptr - offsetof(type,member) );})
#define atomic_read(ref)	(ref.counter)
#define hlist_entry(ptr, type, member)	container_of(ptr, type, member)
#define list_entry(ptr, type, member)  container_of(ptr, type, member)

#define my_hlist_for_each_entry(tpos, pos, head, member, handle) \
				for( pos = (head)->first; \
						 pos && ({ tpos = hlist_entry(pos, typeof(*tpos), member); 1;}); \
						 ({ pos = next_hlist_node(pos, handle); 1;}))

#define my_hlist_for_each_entry_from(tpos, pos, member, handle) \
				for( ; pos && ({tpos = hlist_entry(pos, typeof(*tpos), member); 1;}); \
						({ pos = next_hlist_node(pos, handle); 1;}))

#define my_sk_for_each(__sk, node, list, handle) \
				my_hlist_for_each_entry(__sk, node, list, sk_node, handle)

#define my_inet_twsk_for_each(tw, node, head, handle) \
				my_hlist_for_each_entry(tw, node, head, tw_node, handle)

#define my_sk_for_each_from(__sk, node, handle) \
				if( __sk && ({node = (struct hlist_node *)((unsigned long)__sk + offsetof(struct sock, sk_node)); 1;})) \
					my_hlist_for_each_entry_from(__sk, node, sk_node, handle)

//grobal
struct inet_hashinfo *tcp_hashinfo;
struct inet_ehash_bucket *ehash_list;
int ehash_cnt;

void init_tcp_hashlist(struct sched_struct *handle);
void tcp_hashlist_nextpage(struct sched_struct *handle);
void close_tcp_hashlist(struct sched_struct *handle);
struct hlist_node *next_hlist_node(struct hlist_node *node, struct sched_struct *handle);
struct unix_sock *unix_sk(struct sock *sk, struct sched_struct *handle);
struct tcp_sock *tcp_sk(struct sock *sk, struct sched_struct *handle);
struct inet_sock *inet_sk(struct sock *sk, struct sched_struct *handle);
struct inet_connection_sock *inet_csk(struct sock *sk, struct sched_struct *handle);
//struct tcp_sock *tcp_sk(struct sock *sk);
//struct inet_sock *inet_sk(struct sock *sk);
//struct inet_connection_sock *inet_csk(struct sock *sk);

struct sock *next_unix_socket(struct sock *sk, struct sched_struct *handle);
int unix_show(char *p, struct sock *sock_addr, struct sched_struct *handle);
int hlist_empty(struct hlist_head *h);
struct sock *__sk_head(struct hlist_head *head);
#endif /* __UNIX_H */
