#include "procmem.h"
#include "array.h"
#include "addr.h"
#include <string.h>
#include <stdlib.h>

#define PTE_PSE 0x80

unsigned long translate_proc_address(XCH xc_handle, uint32_t dom,
                                     unsigned long long cr3,
                                     unsigned long long virt)
{
    uint64_t paddr, mask, pte = 0;
    int size, level, pt_levels = 2;
    void *map;
#if 0
    DECLARE_DOMCTL;

    domctl.domain = dom;
    domctl.cmd = XEN_DOMCTL_get_address_size;
    if ( do_domctl(xc_handle, &domctl) != 0 )
        return 0;
    if (domctl.u.address_size.size == 64) {
        pt_levels = 4;
        paddr = cr3 & ~0xfffull;
    } else {
        pt_levels = 3;
        paddr = (uint64_t)xen_cr3_to_pfn(cr3) << PAGE_SHIFT;
    }

    if (pt_levels == 4) {
        virt &= 0x0000ffffffffffffull;
        mask =  0x0000ff8000000000ull;
    } else if (pt_levels == 3) {
        virt &= 0x00000000ffffffffull;
        mask =  0x0000007fc0000000ull;
    } else {
        virt &= 0x00000000ffffffffull;
        mask =  0x00000000ffc00000ull;
    }
    size = (pt_levels == 2 ? 4 : 8);
#else
    /* 64-bit fixed */
    pt_levels = 4;
    paddr = cr3 & ~0xfffull;

    virt &= 0x0000ffffffffffffull;
    mask =  0x0000ff8000000000ull;

    size = 8;
#endif

    /* Walk the pagetables */
    for (level = pt_levels; level > 0; level--) {
        paddr += ((virt & mask) >> (xc_ffs64(mask) - 1)) * size;
        map = xc_map_foreign_range(xc_handle, dom, PAGE_SIZE, PROT_READ, 
                                   paddr >>PAGE_SHIFT);
        if (!map) 
            return 0;
        memcpy(&pte, map + (paddr & (PAGE_SIZE - 1)), size);
        munmap(map, PAGE_SIZE);
        if (!(pte & 1)) 
            return 0;
        paddr = pte & 0x000ffffffffff000ull;
        if (level == 2 && (pte & PTE_PSE)) {
            mask = ((mask ^ ~-mask) >> 1); /* All bits below first set bit */
            return ((paddr & ~mask) | (virt & mask)) >> PAGE_SHIFT;
        }
        mask >>= (pt_levels == 2 ? 10 : 9);
    }
    return paddr >> PAGE_SHIFT;
}

void *proc_addr_trans(uint64_t addr, unsigned int size, pgd_t *pgd,
                      struct sched_struct *handle)
{
    unsigned long pfn, cr3, offset, va;

    pfn = xc_translate_foreign_address(handle->xc_handle, handle->dom, 0,
                                       (unsigned long long)pgd);
    cr3 = pfn << PAGE_SHIFT;

    unsigned long ga;
    xen_pfn_t *arr;
    int n;

    arr = (xen_pfn_t *)malloc(sizeof(xen_pfn_t) * (size / PAGE_SIZE + 1));

    ga = addr & ~(PAGE_SIZE - 1);
    n = 0;

    do {
        pfn = translate_proc_address(handle->xc_handle, handle->dom, cr3,
                                     (unsigned long long)ga);

        arr[n] = pfn;

        ga += PAGE_SIZE;
        n++;
    } while (ga < addr + size);

    va = (unsigned long)xc_map_foreign_batch(handle->xc_handle, handle->dom,
                                             PROT_READ, arr, n);
    if (va == 0) {
        printf("proc_addr_trans: map error (%lx)\n", addr);
        return NULL;
    }

    free(arr);

    offset = (unsigned long)addr & ((1 << PAGE_SHIFT) - 1);

    return (void *)((unsigned long)va + offset);
}

#define PROC_ADDRX(ptr, size, pgd) \
    proc_addr_trans((uint64_t)ptr, size, pgd, handle)

int get_cmdline(char *str, struct task_struct *tk_addr, struct sched_struct *handle)
{
//    struct mm_struct *mm = get_task_mm(task, handle);  /* need to map */
    char *arg, *env;
    int arg_len, env_len;

    struct mm_struct **usr_mm = HREF(tk_addr, mm);
    if (*usr_mm == NULL) {
        return;
    }

    pgd_t **usr_pgd = HREF(*usr_mm, pgd);
    long unsigned int *usr_arg_start = HREF(*usr_mm, arg_start);
    long unsigned int *usr_arg_end = HREF(*usr_mm, arg_end);
    /*
    long unsigned int *usr_env_start = HREF(*usr_mm, env_start);
    long unsigned int *usr_env_end = HREF(*usr_mm, env_end);
    */

    arg_len = *usr_arg_end - *usr_arg_start;
    if (arg_len > PAGE_SIZE)
        arg_len = PAGE_SIZE;

    arg = PROC_ADDRX(*usr_arg_start, arg_len, *usr_pgd);
    memcpy(str, arg, arg_len);

    /*
    env_len = *usr_env_end - *usr_env_start;
    env = PROC_ADDRX(*usr_env_start, env_len, *usr_pgd);
    snprintf(str + arg_len, env_len, env);
    */

    UNMAP(usr_pgd);
    UNMAP(usr_arg_start);
    UNMAP(usr_mm);
    //    UNMAP(usr_env_start);

    return arg_len;
}
