#include "addr.h"
#include "sched.h"
#include "array.h"
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <xenctrl.h>
#include <stdio.h>
#include <inttypes.h>
#include <sys/stat.h>

int
sched_init(int dom, struct sched_struct* handle)
{
  handle->dom = dom;
  handle->cpu = 0;

  // xen handler
#ifdef XEN41
  if ((handle->xc_handle = xc_interface_open(0, 0, 0)) < 0) {
#else
  if ((handle->xc_handle = xc_interface_open()) < 0) {
#endif
    perror("xc_interface_open");
    return -1;
  }

  // pause domain
  if (xc_domain_pause(handle->xc_handle, handle->dom) < 0) {
    perror("xc_domain_pause");
    sched_exit(handle);
    return -1;
  }

  handle->init_task = (struct task_struct*)INIT_TASK_ADDR;

  return 0;
}

int
sched_exit(struct sched_struct* handle)
{
  int ret = 0;
  if (xc_domain_unpause(handle->xc_handle, handle->dom) < 0) {
    perror("xc_domain_unpause");
    ret = -1;
  }
  if (xc_interface_close(handle->xc_handle) < 0) {
    perror("close");
    ret = -1;
  }

#if 0
  if (sigprocmask(SIG_SETMASK, &handle->sig_saved, NULL) < 0) {
    perror("sigprocmask");
    ret = -1;
  }
#endif

  memset(handle, 0, sizeof(struct sched_struct));

  return ret;
}

struct task_struct* 
next_task(struct task_struct *task, struct sched_struct *handle)
{
  struct task_struct *ptr;
  struct list_head **usr_tasks_next = HREF(task, tasks.next);

  ptr = (struct task_struct*)((unsigned long)(*usr_tasks_next) -
			      (offsetof(struct task_struct, tasks.next)));

//  UNMAP(usr_tasks_next);

  return ptr;
}

struct task_struct *get_task_struct(struct task_struct *task, struct sched_struct *handle)
{
    struct task_struct *usr_task = ADDRX(task);

    return usr_task;
}
