#define FUSE_USE_VERSION  26
   
#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <stdlib.h>

#include "linux_types.h"
#include "sched.h"
#include "addr.h"
#include "array.h"
#include "net.h"
#include "procmem.h"

#include <sys/time.h>
#include <sys/xattr.h>

#define MAX_PATH 256
#define MAX_NAME 64
#define MAX_DATA 100000

struct fuse_node {
  char name[MAX_NAME];
  char *data;
	struct stat st;
  
  struct fuse_node *down;
  struct fuse_node *next;
};

struct fuse_node *top = NULL;
struct fuse_node *bottom = NULL;
struct sched_struct handle_st;
struct sched_struct *handle = &handle_st;
int dirs = 0;

void save_log(const char *file_name, const char *str)
{
  FILE *fp;
  char path[MAX_PATH] = "/home/yone/";
  strcat(path, file_name);
  fp = fopen(path, "a");
  fprintf(fp, "%s\n", str);
  fclose(fp);
}

struct fuse_node *last_next_node(struct fuse_node *node)
{
	if(node->next == NULL)
		return node;
	else
		return last_next_node(node->next);
}

void delete_all_node(struct fuse_node *node)
{
	struct fuse_node *next, *down;

	if(node == NULL)
		return;

	next = node->next;
	down = node->down;

	if(node->data != NULL)
		free(node->data);

	delete_all_node(next);
	delete_all_node(down);

}

void separate_str(const char *str, char *path1, char *path2)
{
  char *p, ori[MAX_PATH];
  strcpy(ori, str);
  p = strchr(ori, '/');
  if(p != NULL){
    strcpy(path2, p);
    strcpy(p, "");
  }
  strcpy(path1, ori);
}

struct fuse_node *get_node(struct fuse_node *head, const char *path)
{
	struct fuse_node *node = head;
	char path1[MAX_PATH], path2[MAX_PATH] = "";
	int chk = 0;

	separate_str(&path[1], path1, path2);

	while(node){
		if(strcmp(node->name, path1) == 0){
			if(strcmp(path2, "") != 0){
				return get_node(node->down, path2);
			} else {
				return node;
			}
		}
		node = node->next;
	}
	return NULL;
}

struct fuse_node *find_node(const char *path)
{
	return get_node(top->down, path);
}

struct fuse_node *get_empty_node(struct fuse_node *node)
{
	if(node->down == NULL){
		node->down = (struct fuse_node *)malloc(sizeof(struct fuse_node));
		node->down->next = NULL;
		node->down->down = NULL;
		node->down->data = NULL;
		return node->down;
	} else { 
		struct fuse_node *last_node = last_next_node(node->down);
		last_node->next = (struct fuse_node *)malloc(sizeof(struct fuse_node));
		last_node->next->next = NULL;
		last_node->next->down = NULL;
		last_node->next->data = NULL;
		return last_node->next;
	}
}

void print_data(struct fuse_node *fnode, char *data, int size)
{
	fnode->st.st_size = size;
	fnode->data = (char *)malloc(size);
	memcpy(fnode->data, data, size);
}

static inline struct fuse_node *get_fnode(struct fuse_file_info *fi)
{
	return (struct fuse_node *)(uintptr_t)fi->fh;
}

struct fuse_node *add_node(struct fuse_node *fnode, const char *name, mode_t mode, nlink_t nlink)
{
	struct fuse_node *new_node = get_empty_node(fnode);
	new_node->st.st_mode = mode;
	new_node->st.st_nlink = nlink;
	strcpy(new_node->name, name);
	return new_node;
}                 

void init_shadow_proc_dirs(int dom)
{
	struct task_struct *task,*tk_addr;
	int fd;
	char str[MAX_DATA];
	int i;
	struct fuse_node *fnode, *new_node;
        int size;

	sched_init(dom, handle);
	tk_addr = next_task(handle->init_task, handle);

	task = get_task_struct(tk_addr, handle);

	struct uts_namespace **user_ptr_uts_ns = HREF(task->nsproxy, uts_ns);
	struct new_utsname *usr_name = HREF(*user_ptr_uts_ns, name);

	sprintf(str, "%s\n%s\n%s\n%s\n%s\n%s\n", usr_name->sysname, usr_name->nodename, usr_name->release, usr_name->version, usr_name->machine, usr_name->domainname);
	fnode = add_node(top, ".uname", S_IFREG | 0444, 1);
	fnode->st.st_size = strlen(str);
	fnode->data = (char *)malloc(sizeof(char)*(fnode->st.st_size + 1));
	strcpy(fnode->data, str);

        char *linux_proc_banner = ADDRX2(LINUX_PROC_BANNER, 256);

	fnode = add_node(top, "version", S_IFREG | 0444, 1);
	sprintf(str, linux_proc_banner,
                usr_name->sysname, usr_name->release, usr_name->version);
	fnode->st.st_size = strlen(str);
	fnode->data = (char *)malloc(sizeof(char)*(fnode->st.st_size + 1));
	strcpy(fnode->data, str);

	do{
		task = get_task_struct(tk_addr, handle);
		sprintf(str, "%d", task->pid);
		fnode = add_node(top, str, S_IFDIR | 0555, 2);
		struct fuse_node *task_node = add_node(fnode, "task", S_IFDIR | 0555, 2);
		struct fuse_node *task_node_dir = add_node(task_node, str, S_IFDIR | 0555, 2);
		new_node = add_node(fnode, "stat", S_IFREG | 0444, 1);
		proc_tid_stat(str, tk_addr, handle);
		print_data(new_node, str, strlen(str));
		new_node = add_node(task_node_dir, "stat", S_IFREG | 0444, 1);
		print_data(task_node, str, strlen(str));

		new_node = add_node(fnode, "status", S_IFREG | 0444, 1);
		proc_pid_status(str, tk_addr, handle);
		print_data(new_node, str, strlen(str));
		new_node = add_node(task_node_dir, "status", S_IFREG | 0444, 1);
		print_data(new_node, str, strlen(str));

		new_node = add_node(fnode, "cmdline", S_IFREG | 0444, 1);
		size = get_cmdline(str, tk_addr, handle);
		print_data(new_node, str, size);
		new_node = add_node(task_node_dir, "cmdline", S_IFREG | 0444, 1);
		print_data(new_node, str, size);

		struct task_struct *next_th = next_thread(tk_addr, handle);
		while(next_th != tk_addr){
			task = get_task_struct(next_th, handle);
			sprintf(str, "%d", task->pid);
			task_node_dir = add_node(task_node, str, S_IFDIR | 0555, 2);
			new_node = add_node(task_node_dir, "stat", S_IFREG | 0444, 1);
			proc_tid_stat(str, tk_addr, handle);
			print_data(new_node, str, strlen(str));

			new_node = add_node(task_node_dir, "status", S_IFREG | 0444, 1);
			proc_pid_status(str, tk_addr, handle);
			print_data(new_node, str, strlen(str));

			new_node = add_node(task_node_dir, "cmdline", S_IFREG | 0444, 1);
			get_cmdline(str, tk_addr, handle);
			print_data(new_node, str, strlen(str));

			next_th = next_thread(next_th, handle);	
		}

	
		tk_addr = next_task(tk_addr, handle);
		dirs++;
	}while(tk_addr != handle->init_task);

	fnode = add_node(top, "stat", S_IFREG | 0444, 1);
	fd = open("/proc/stat", O_RDONLY);
	if(fd == -1){
		perror("open:/proc/stat");
	}else{
		read(fd, str, MAX_DATA);
		print_data(fnode, str, strlen(str));
		dirs++;
	}
	close(fd);
	fnode = add_node(top, "self", S_IFLNK | 0777, 1);
	fnode = add_node(top, "uptime", S_IFLNK | 0777, 1);
	fnode = add_node(top, "mounts", S_IFLNK | 0777, 1);
	fnode = add_node(top, "filesystems", S_IFLNK | 0777, 1);
/*
	fnode = add_node(top, "self", S_IFDIR | 0555, 1);
	new_node = add_node(fnode, "stat", S_IFREG | 0444, 1);
	fd = open("/proc/self/stat", O_RDONLY);
	if(fd == -1){
		perror("fopen:/proc/self/stat");
	}else{
		read(fd, str, MAX_DATA);
		print_data(new_node, str);
		dirs++;
	}
	close(fd);
*/
	new_node = add_node(fnode, "task", S_IFDIR | 0555, 2);

	fd = open("/proc/meminfo", O_RDONLY);
	if(fd == -1){
		perror("open:/proc/meminfo");
	}else{
		fnode = add_node(top, "meminfo", S_IFREG | 0444, 1);
		read(fd, str, MAX_DATA);
		print_data(fnode, str, strlen(str));
		dirs++;
	}
	close(fd);

	fnode = add_node(top, "sys", S_IFDIR | 0555, 2);
	new_node = add_node(fnode, "kernel", S_IFDIR | 0555, 2);
	new_node = add_node(new_node, "pid_max", S_IFREG | 0444, 1);
	sprintf(str, "%lu\n", PAGE_SIZE * 8);
	print_data(new_node, str, strlen(str));
	dirs++;

	fnode = add_node(top, "net", S_IFDIR | 0555, 2);
	new_node = add_node(fnode, "unix", S_IFREG | 0444, 1);
	sprintf(str, "");
	for(i = 0; i <= UNIX_HASH_SIZE; i++){
		struct hlist_head *p_list = (struct hlist_head *)((unsigned long)UNIX_SOCKET_TABLE + i*8);
		struct hlist_head *user_list = ADDRX(p_list);
		if(!hlist_empty(user_list)){
			struct sock *tmp_sock = __sk_head(user_list);
			unix_show(str, tmp_sock, handle);
		}
	}
	print_data(new_node, str, strlen(str));

	new_node = add_node(fnode, "udp", S_IFREG | 0444, 1);
	sprintf(str, "");
	for(i = 0; i <= UDP_HTABLE_SIZE; i++){
		struct hlist_head *p_list = (struct hlist_head *)((unsigned long)UDP_HASH + i*8);
		struct hlist_head *udp_list = ADDRX(p_list);
		if(!hlist_empty(udp_list)){
			struct sock *udp_sock = __sk_head(udp_list);
			udp_show(str, udp_sock, handle, &i);
		}
	}
	print_data(new_node, str, strlen(str));

	new_node = add_node(fnode, "tcp", S_IFREG | 0444, 1);
	sprintf(str, "");
	tcp_open(str, handle);
	print_data(new_node, str, strlen(str));

	dirs++;

        /*
	fnode = add_node(top, "knark", S_IFDIR | 0555, 2);
	dirs++;
        */

	sched_exit(handle);
}

void init_shadow_proc(int domid)
{
	sched_init(domid, handle);
	sched_exit(handle);

	top = (struct fuse_node *)malloc(sizeof(struct fuse_node));
        top->next = NULL;
        top->down = NULL;
        top->data = NULL;
	sprintf(top->name, "/");

	init_shadow_proc_dirs(domid);
}

static int shadow_proc_getattr(const char *path, struct stat *stbuf)
{
	int res = 0;

	struct fuse_node *fnode = find_node(path); 
	if(strcmp(path, "/") == 0){
		memset(stbuf, 0, sizeof(struct stat));
		stbuf->st_mode = S_IFDIR | 0555;
		stbuf->st_nlink = 2;
	} else if(fnode != NULL){
		memset(stbuf, 0, sizeof(struct stat));
		stbuf->st_mode = fnode->st.st_mode; 
		stbuf->st_nlink = fnode->st.st_nlink;
		stbuf->st_size = fnode->st.st_size;
	} else 
		res = -ENOENT;

	return res;
}

static int shadow_proc_readlink(const char *path, char *buf, size_t size)
{
	int res;
	struct stat stbuf;

	sprintf(buf, "/proc%s", path);
	res = lstat(buf, &stbuf);
	if(res == -1)
		return -ENOENT;
		
	return 0;
}

static int shadow_proc_opendir(const char *path, struct fuse_file_info *fi)
{
	int res = 0;

	if(!strcmp(path, "/")){
		fi->fh = (uintptr_t)top->down;
	} else {
		struct fuse_node *fnode = find_node(path);
		if(fnode != NULL)
			fi->fh = (uintptr_t)fnode->down;
		else
			return -ENOENT;	
	}
	return 0;
}

static int shadow_proc_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi)
{
	struct fuse_node *fnode = get_fnode(fi);      

	char str[256];

	if(fnode == NULL)
		return -ENOENT;

	filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);

	while(fnode != NULL){
		filler(buf, fnode->name,  &(fnode->st), 0);
		fnode = fnode->next;
	}

	return 0;
}


static int shadow_proc_open(const char *path, struct fuse_file_info *fi)
{
  struct fuse_node *fnode = find_node(path);
  if(fnode == NULL)
    return -ENOENT;

  fi->fh = (uintptr_t)fnode;

  return 0;
}

static int shadow_proc_read(const char *path, char *buf, size_t size, off_t offset,
		     struct fuse_file_info *fi)
{
	struct fuse_node *fnode = get_fnode(fi);

	if(fnode == NULL)
		return -EINVAL;

	if(offset > fnode->st.st_size)
		return -EINVAL;

	if(offset + size > fnode->st.st_size)
		size = fnode->st.st_size - offset;

	memcpy(buf, &(fnode->data[offset]), size);

	return size;

}

static struct fuse_operations proc_oper = {
  .getattr   = shadow_proc_getattr,
	.readlink	= shadow_proc_readlink,
	.opendir = shadow_proc_opendir,
  .readdir = shadow_proc_readdir,
  .open   = shadow_proc_open,
  .read   = shadow_proc_read,
};

int main(int argc, char *argv[])
{
	int domid;
  if(getuid() != 0){
    fprintf(stderr, "need root privileges\n");
    return 1;
  }
	if(argc < 3){
		fprintf(stderr, "input domU's ID\n");
		return 1;
	}
	domid = atoi(argv[argc-1]);
	argc--;	
  init_shadow_proc(domid);

  return fuse_main(argc, argv, &proc_oper, NULL);
}
